import json
import os
from argparse import ArgumentParser
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import List

import dacite
import matplotlib.pyplot as plt
import numpy as np


@dataclass
class DataPoint:
    num_teleport: int
    num_local: int
    tel_succ_prob: float
    tel_succ_prob_lower: float
    tel_succ_prob_upper: float
    loc_succ_prob: float
    loc_succ_prob_lower: float
    loc_succ_prob_upper: float
    makespan: float


@dataclass
class DataMeta:
    timestamp: str
    num_teleport: int
    num_local: int
    num_runs: int
    t1: float
    t2: float
    latency_factor: float
    cc_latency: float
    local_busy_duration: float
    network_bin_len: int
    network_period: int
    network_first_bin: int
    num_qubits_bob: int


@dataclass
class Data:
    meta: DataMeta
    data_points: List[DataPoint]


def relative_to_cwd(file: str) -> str:
    return os.path.join(os.path.dirname(__file__), file)


def create_png(filename: str):
    output_dir = relative_to_cwd("plots")
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    output_path = os.path.join(output_dir, f"{filename}.png")
    plt.savefig(output_path)
    print(f"plot written to {output_path}")


def create_meta(filename: str, datas: List[Data], plot_tel: bool):
    output_dir = relative_to_cwd("plots")
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    output_path = os.path.join(output_dir, f"{filename}.json")
    meta = {}
    meta["datafiles"] = [
        {
            "num_teleport": data.meta.num_teleport,
            "num_local": data.meta.num_local,
            "timestamp": data.meta.timestamp,
        }
        for data in datas
    ]
    if plot_tel:
        meta["plotted_succ_prob"] = "teleport"
    else:
        meta["plotted_succ_prob"] = "local"
    with open(output_path, "w") as metafile:
        json.dump(meta, metafile)


def load_data(path: str) -> Data:
    with open(relative_to_cwd(path), "r") as f:
        all_data = json.load(f)

    # assert isinstance(all_data, list)
    # return [dacite.from_dict(DataPoint, entry) for entry in all_data]
    return dacite.from_dict(Data, all_data)


def plot_heatmap(
    timestamp: str, datas: List[Data], num_teleport: int, num_local: int, plot_tel: bool
) -> None:
    fig, ax = plt.subplots()

    ax.grid()
    ax.set_xlabel("Number of teleportation instances", fontsize=14)
    ax.set_ylabel("Number of local instances", fontsize=14)

    plot_data = np.empty((num_local, num_teleport))

    for data in datas:
        tel = data.meta.num_teleport
        loc = data.meta.num_local
        tel_succ = data.data_points[0].tel_succ_prob
        loc_succ = data.data_points[0].loc_succ_prob
        if plot_tel:
            plot_data[loc - 1][tel - 1] = tel_succ
        else:
            plot_data[loc - 1][tel - 1] = loc_succ

    plt.pcolor(plot_data, cmap="viridis")
    cbar = plt.colorbar()
    cbar.ax.tick_params(labelsize=12)

    if plot_tel:
        ax.set_title("Teleportation success probability", wrap=True, fontsize=14)
    else:
        ax.set_title("Local success probability", wrap=True, fontsize=14)

    ax.set_xticks(np.arange(0.5, num_teleport + 0.5), range(1, num_teleport + 1))
    ax.set_yticks(np.arange(0.5, num_local + 0.5), range(1, num_local + 1))

    # ax.set_ylim(0.75, 0.9)
    # ax.legend(loc="upper left")

    tel_or_loc = "tel" if plot_tel else "loc"
    create_png(f"LAST_{tel_or_loc}")
    create_png(f"{timestamp}_{tel_or_loc}")


def heat_map(num_teleport: int, num_local: int):
    datas: List[Data] = []
    for tel in range(1, num_teleport + 1):
        for loc in range(1, num_local + 1):
            data = load_data(f"data/sweep_teleport_local_{tel}_{loc}/LAST.json")
            datas.append(data)
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")

    # Plot Teleportation success probability
    create_meta("LAST_tel_meta", datas, plot_tel=True)
    create_meta(f"{timestamp}_tel_meta", datas, plot_tel=True)
    plot_heatmap(timestamp, datas, num_teleport, num_local, plot_tel=True)

    # Plot Local success probability
    create_meta("LAST_loc_meta", datas, plot_tel=False)
    create_meta(f"{timestamp}_loc_meta", datas, plot_tel=False)
    plot_heatmap(timestamp, datas, num_teleport, num_local, plot_tel=False)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--num_teleport", "-t", type=int, required=True)
    parser.add_argument("--num_local", "-l", type=int, required=True)

    args = parser.parse_args()

    num_teleport = args.num_teleport
    num_local = args.num_local

    heat_map(num_teleport, num_local)



from typing import Optional

from qoala.lang.ehi import EhiNetworkInfo, UnitModule
from qoala.lang.program import QoalaProgram
from qoala.runtime.lhi import LhiLatencies, LhiTopologyBuilder
from qoala.runtime.lhi_to_ehi import LhiConverter
from qoala.runtime.ntf import GenericNtf
from qoala.runtime.program import ProgramInput, ProgramInstance
from qoala.sim.build import build_qprocessor_from_topology
from qoala.sim.procnode import ProcNode


class ObjectBuilder:
    @classmethod
    def simple_procnode(cls, name: str, num_qubits: int) -> ProcNode:
        network_ehi = EhiNetworkInfo(nodes={0: name}, links={})
        topology = LhiTopologyBuilder.perfect_uniform_default_gates(num_qubits)
        qprocessor = build_qprocessor_from_topology(f"{name}_processor", topology)
        return ProcNode(
            name=name,
            qprocessor=qprocessor,
            qdevice_topology=topology,
            latencies=LhiLatencies.all_zero(),
            ntf_interface=GenericNtf(),
            network_ehi=network_ehi,
        )

    @classmethod
    def simple_program_instance(
        cls, program: QoalaProgram, pid: int = 0, inputs: Optional[ProgramInput] = None
    ) -> ProgramInstance:
        topology = LhiTopologyBuilder.perfect_uniform_default_gates(1)
        ehi = LhiConverter.to_ehi(topology, GenericNtf())
        unit_module = UnitModule.from_full_ehi(ehi)

        if inputs is None:
            inputs = ProgramInput.empty()

        return ProgramInstance(
            pid,
            program,
            inputs,
            unit_module=unit_module,
        )



from typing import Any, Dict, FrozenSet, Optional, Type

from netsquid.components.models.qerrormodels import (
    DepolarNoiseModel,
    QuantumErrorModel,
    T1T2NoiseModel,
)
from netsquid_magic.state_delivery_sampler import (
    DepolariseWithFailureStateSamplerFactory,
    PerfectStateSamplerFactory,
)

from qoala.lang.ehi import (
    EhiGateInfo,
    EhiLatencies,
    EhiLinkInfo,
    EhiNetworkInfo,
    EhiNetworkSchedule,
    EhiNetworkTimebin,
    EhiNodeInfo,
    EhiQubitInfo,
)
from qoala.runtime.lhi import (
    LhiGateInfo,
    LhiLatencies,
    LhiLinkInfo,
    LhiNetworkInfo,
    LhiNetworkSchedule,
    LhiNetworkTimebin,
    LhiQubitInfo,
    LhiTopology,
)
from qoala.runtime.ntf import NtfInterface
from qoala.util.math import prob_max_mixed_to_fidelity


class LhiConverter:
    @classmethod
    def error_model_to_rate(
        cls, model: Type[QuantumErrorModel], model_kwargs: Dict[str, Any]
    ) -> float:
        if model == DepolarNoiseModel:
            return model_kwargs["depolar_rate"]  # type: ignore
        elif model == T1T2NoiseModel:
            # TODO use T2 somehow
            return model_kwargs["T1"]  # type: ignore
        else:
            raise RuntimeError("Unsupported LHI Error model")

    @classmethod
    def qubit_info_to_ehi(cls, info: LhiQubitInfo) -> EhiQubitInfo:
        return EhiQubitInfo(
            is_communication=info.is_communication,
            decoherence_rate=cls.error_model_to_rate(
                info.error_model, info.error_model_kwargs
            ),
        )

    @classmethod
    def gate_info_to_ehi(cls, info: LhiGateInfo, ntf: NtfInterface) -> EhiGateInfo:
        # TODO: deal with mapping to multiple gates
        instr = ntf.native_to_netqasm(info.instruction)[0]  # (!)
        duration = info.duration
        decoherence = cls.error_model_to_rate(info.error_model, info.error_model_kwargs)
        return EhiGateInfo(
            instruction=instr, duration=duration, decoherence=decoherence
        )

    @classmethod
    def to_ehi(
        cls,
        topology: LhiTopology,
        ntf: NtfInterface,
        latencies: Optional[LhiLatencies] = None,
    ) -> EhiNodeInfo:
        if latencies is None:
            latencies = LhiLatencies.all_zero()

        qubit_infos = {
            id: cls.qubit_info_to_ehi(qi) for (id, qi) in topology.qubit_infos.items()
        }
        single_gate_infos = {
            id: [cls.gate_info_to_ehi(gi, ntf) for gi in gis]
            for (id, gis) in topology.single_gate_infos.items()
        }
        multi_gate_infos = {
            ids: [cls.gate_info_to_ehi(gi, ntf) for gi in gis]
            for (ids, gis) in topology.multi_gate_infos.items()
        }
        if topology.all_qubit_gate_infos is None:
            all_qubit_gate_infos = []
        else:
            all_qubit_gate_infos = [
                cls.gate_info_to_ehi(gi, ntf) for gi in topology.all_qubit_gate_infos
            ]

        flavour = ntf.flavour()

        ehi_latencies = EhiLatencies(
            latencies.host_instr_time,
            latencies.qnos_instr_time,
            latencies.host_peer_latency,
            latencies.internal_sched_latency,
        )

        return EhiNodeInfo(
            qubit_infos=qubit_infos,
            flavour=flavour,
            single_gate_infos=single_gate_infos,
            multi_gate_infos=multi_gate_infos,
            all_qubit_gate_infos=all_qubit_gate_infos,
            latencies=ehi_latencies,
        )

    @classmethod
    def link_info_to_ehi(cls, info: LhiLinkInfo) -> EhiLinkInfo:
        if info.sampler_factory == PerfectStateSamplerFactory:
            return EhiLinkInfo(duration=info.state_delay, fidelity=1.0)
        elif info.sampler_factory == DepolariseWithFailureStateSamplerFactory:
            expected_gen_duration = (
                info.sampler_kwargs["cycle_time"] / info.sampler_kwargs["prob_success"]
            )
            duration = expected_gen_duration + info.state_delay
            fidelity = prob_max_mixed_to_fidelity(
                2, info.sampler_kwargs["prob_max_mixed"]
            )
            return EhiLinkInfo(duration=duration, fidelity=fidelity)
        else:
            raise NotImplementedError

    @classmethod
    def network_to_ehi(cls, info: LhiNetworkInfo) -> EhiNetworkInfo:
        links: Dict[FrozenSet[int], EhiLinkInfo] = {}
        for (node_link, link_info) in info.links.items():
            ehi_link = cls.link_info_to_ehi(link_info)
            links[node_link] = ehi_link
        return EhiNetworkInfo(info.nodes, links)

    @classmethod
    def timebin_to_ehi(cls, bin: LhiNetworkTimebin) -> EhiNetworkTimebin:
        return EhiNetworkTimebin(bin.nodes, bin.pids)

    @classmethod
    def netschedule_to_ehi(cls, schedule: LhiNetworkSchedule) -> EhiNetworkSchedule:
        return EhiNetworkSchedule(
            bin_length=schedule.bin_length,
            first_bin=schedule.first_bin,
            bin_pattern=[cls.timebin_to_ehi(bin) for bin in schedule.bin_pattern],
            repeat_period=schedule.repeat_period,
        )



from __future__ import annotations

import datetime
import json
import os
import time
from argparse import ArgumentParser
from dataclasses import asdict, dataclass
from math import ceil
from pathlib import Path
from typing import List

import netsquid as ns

from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    ClassicalConnectionConfig,
    LatenciesConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import ProgramInput
from qoala.util.runner import AppResult, run_two_node_app


def relative_to_cwd(file: str) -> str:
    return os.path.join(os.path.dirname(__file__), file)


def create_procnode_cfg(
    name: str, id: int, num_qubits: int, determ: bool
) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=TopologyConfig.perfect_config_uniform_default_params(num_qubits),
        latencies=LatenciesConfig(qnos_instr_time=1000),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
        determ_sched=determ,
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


def run_teleport(
    num_iterations: int,
    linear: bool,
    num_qubits_alice: int,
    num_qubits_bob: int,
    cc: float,
) -> AppResult:
    ns.sim_reset()

    alice_id = 1
    bob_id = 0

    alice_node_cfg = create_procnode_cfg(
        "alice", alice_id, num_qubits_alice, determ=True
    )
    bob_node_cfg = create_procnode_cfg("bob", bob_id, num_qubits_bob, determ=True)
    bob_node_cfg.topology.qubits[0].qubit_config.noise_config.to_error_model_kwargs()[
        "T2"
    ] = 1

    cconn = ClassicalConnectionConfig.from_nodes(alice_id, bob_id, cc)
    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[alice_node_cfg, bob_node_cfg], link_duration=1000
    )
    network_cfg.cconns = [cconn]

    alice_program = load_program("teleport_alice.iqoala")
    bob_program = load_program("teleport_bob.iqoala")

    # state = 5 -> teleport |1> state
    alice_input = ProgramInput({"bob_id": bob_id, "state": 5})
    # state = 4 -> measure in +Z, i.e. expect a "1" outcome
    bob_input = ProgramInput({"alice_id": alice_id, "state": 4})

    app_result = run_two_node_app(
        num_iterations=num_iterations,
        programs={"alice": alice_program, "bob": bob_program},
        program_inputs={"alice": alice_input, "bob": bob_input},
        network_cfg=network_cfg,
        linear_for={"alice": True, "bob": linear},
    )
    # print(f"makespan: {app_result.total_duration:_}")

    return app_result


def get_teleport_makespan(
    num_iterations: int,
    linear: bool,
    num_qubits_alice: int,
    num_qubits_bob: int,
    cc: float,
) -> float:
    result = run_teleport(
        num_iterations=num_iterations,
        linear=linear,
        num_qubits_alice=num_qubits_alice,
        num_qubits_bob=num_qubits_bob,
        cc=cc,
    )

    bob_result = result.batch_results["bob"]

    program_results = bob_result.results
    outcomes = [result.values["outcome"] for result in program_results]
    # print(outcomes)
    assert all(outcome == 1 for outcome in outcomes)

    return result.total_duration


@dataclass
class DataPoint:
    linear: bool
    num_qubits_bob: int
    makespan: float
    approx_seq: float
    approx_int: float


@dataclass
class DataMeta:
    timestamp: str
    num_iterations: int
    num_qubits_alice: int
    cc: float
    sim_duration: float


@dataclass
class Data:
    meta: DataMeta
    data_points: List[DataPoint]


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--num_iterations", "-n", type=int, required=True)

    args = parser.parse_args()
    num_iterations = args.num_iterations

    num_qubits_alice = 2
    cc = 1e7

    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    start = time.time()
    data_points: List[DataPoint] = []

    for linear in [False, True]:
        for num_qubits_bob in range(1, 6):
            print(f"{linear}, {num_qubits_alice}, {num_qubits_bob}")
            makespan = get_teleport_makespan(
                num_iterations, linear, num_qubits_alice, num_qubits_bob, cc
            )
            approx_seq = num_iterations * cc
            approx_int = ceil(num_iterations / num_qubits_bob) * cc
            point = DataPoint(linear, num_qubits_bob, makespan, approx_seq, approx_int)
            data_points.append(point)

    end = time.time()
    duration = round(end - start, 2)

    abs_dir = relative_to_cwd(f"data")
    Path(abs_dir).mkdir(parents=True, exist_ok=True)
    last_path = os.path.join(abs_dir, "LAST.json")
    timestamp_path = os.path.join(abs_dir, f"{timestamp}.json")

    meta = DataMeta(timestamp, num_iterations, num_qubits_alice, cc, duration)
    data = Data(meta=meta, data_points=data_points)
    json_data = asdict(data)

    # for p in data:
    #     print(f"{p.linear}, {p.num_qubits_alice}, {p.num_qubits_bob}, {p.makespan:_}")

    with open(last_path, "w") as datafile:
        json.dump(json_data, datafile)
    with open(timestamp_path, "w") as datafile:
        json.dump(json_data, datafile)



import os
import random
from copy import deepcopy
from dataclasses import dataclass
from enum import Enum, auto
from typing import Dict, List, Optional

import netsquid as ns

from qoala.lang.ehi import UnitModule
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import ProcNodeNetworkConfig  # type: ignore
from qoala.runtime.program import BatchInfo, BatchResult, ProgramBatch, ProgramInput
from qoala.runtime.statistics import SchedulerStatistics
from qoala.runtime.task import TaskGraph, TaskGraphBuilder
from qoala.sim.build import build_network_from_config


class SchedulerType(Enum):
    NO_SCHED = 0
    FCFS = auto()
    QOALA = auto()


@dataclass
class AppResult:
    batch_results: Dict[str, BatchResult]
    statistics: Dict[str, SchedulerStatistics]
    total_duration: float


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


def create_batch(
    program: QoalaProgram,
    unit_module: UnitModule,
    inputs: List[ProgramInput],
    num_iterations: int,
) -> BatchInfo:
    return BatchInfo(
        program=program,
        unit_module=unit_module,
        inputs=inputs,
        num_iterations=num_iterations,
        deadline=0,
    )


def run_two_node_app_separate_inputs(
    num_iterations: int,
    programs: Dict[str, QoalaProgram],
    program_inputs: Dict[str, List[ProgramInput]],
    network_cfg: ProcNodeNetworkConfig,
    linear: bool = False,
    linear_for: Optional[Dict[str, bool]] = None,
) -> AppResult:
    ns.sim_reset()
    ns.set_qstate_formalism(ns.QFormalism.DM)
    seed = random.randint(0, 1000)
    ns.set_random_state(seed=seed)

    network = build_network_from_config(network_cfg)

    names = list(programs.keys())
    assert len(names) == 2
    other_name = {names[0]: names[1], names[1]: names[0]}
    batches: Dict[str, ProgramBatch] = {}  # node -> batch

    for name in names:
        procnode = network.nodes[name]
        program = programs[name]
        inputs = program_inputs[name]

        unit_module = UnitModule.from_full_ehi(procnode.memmgr.get_ehi())
        batch_info = create_batch(program, unit_module, inputs, num_iterations)
        batches[name] = procnode.submit_batch(batch_info)

    for name in names:
        procnode = network.nodes[name]

        remote_batch = batches[other_name[name]]
        remote_pids = {remote_batch.batch_id: [p.pid for p in remote_batch.instances]}

        if linear_for is not None:
            procnode.initialize_processes(remote_pids, linear=linear_for[name])
        else:
            procnode.initialize_processes(remote_pids, linear=linear)

        # tasks = procnode.scheduler.get_tasks_to_schedule()
        # if linear:
        #     merged = TaskGraphBuilder.merge_linear(tasks)
        # else:
        #     if linear_for[name]:
        #         merged = TaskGraphBuilder.merge_linear(tasks)
        #     else:
        #         merged = TaskGraphBuilder.merge(tasks)
        # procnode.scheduler.upload_task_graph(merged)

        # logger = LogManager.get_stack_logger()
        # for batch_id, prog_batch in procnode.scheduler.get_batches().items():
        #     task_graph = prog_batch.instances[0].task_graph
        #     num = len(prog_batch.instances)
        #     logger.info(f"batch {batch_id}: {num} instances each with task graph:")
        #     logger.info(task_graph)

    network.start()
    ns.sim_run()

    results: Dict[str, BatchResult] = {}
    statistics: Dict[str, SchedulerStatistics] = {}

    for name in names:
        procnode = network.nodes[name]
        # only one batch (ID = 0), so get value at index 0
        results[name] = procnode.scheduler.get_batch_results()[0]
        statistics[name] = procnode.scheduler.get_statistics()

    total_duration = ns.sim_time()
    return AppResult(results, statistics, total_duration)


def run_two_node_app_separate_inputs_plus_constant_tasks(
    num_iterations: int,
    num_const_tasks: int,
    node1: str,
    node2: str,
    prog_node1: QoalaProgram,
    prog_node1_inputs: List[ProgramInput],
    prog_node2: QoalaProgram,
    prog_node2_inputs: List[ProgramInput],
    const_prog_node2: QoalaProgram,
    const_prog_node2_inputs: List[ProgramInput],
    const_period: int,
    const_start: int,
    network_cfg: ProcNodeNetworkConfig,
    sched_typ: SchedulerType,
    linear: bool = False,
) -> AppResult:
    ns.sim_reset()
    ns.set_qstate_formalism(ns.QFormalism.DM)
    seed = random.randint(0, 1000)
    ns.set_random_state(seed=seed)

    network = build_network_from_config(network_cfg)

    procnode1 = network.nodes[node1]
    unit_module1 = UnitModule.from_full_ehi(procnode1.memmgr.get_ehi())
    batch_info1 = create_batch(
        prog_node1, unit_module1, prog_node1_inputs, num_iterations
    )
    batch_node1 = procnode1.submit_batch(batch_info1)

    procnode2 = network.nodes[node2]
    unit_module2 = UnitModule.from_full_ehi(procnode2.memmgr.get_ehi())
    batch_info2 = create_batch(
        prog_node2, unit_module2, prog_node2_inputs, num_iterations
    )
    batch_node2 = procnode2.submit_batch(batch_info2)

    batch_info_const = create_batch(
        const_prog_node2, unit_module2, const_prog_node2_inputs, num_const_tasks
    )
    batch_const = procnode2.submit_const_batch(batch_info_const)

    # Init

    remote_pids1 = {batch_node2.batch_id: [p.pid for p in batch_node2.instances]}
    procnode1.initialize_processes(remote_pids1)

    remote_pids2 = {batch_node1.batch_id: [p.pid for p in batch_node1.instances]}
    procnode2.initialize_processes(remote_pids2)

    # Upload tasks

    # tasks1 = procnode1.scheduler.get_tasks_to_schedule()
    # if linear:
    #     merged1 = TaskGraphBuilder.merge_linear(tasks1)
    # else:
    #     merged1 = TaskGraphBuilder.merge(tasks1)
    # procnode1.scheduler.upload_task_graph(merged1)

    # tasks2 = procnode2.scheduler.get_tasks_to_schedule_for(batch_node2.batch_id)
    # if linear:
    #     merged2 = TaskGraphBuilder.merge_linear(tasks2)
    # else:
    #     merged2 = TaskGraphBuilder.merge(tasks2)

    tasks_const: List[TaskGraph] = []
    task_counter = 0
    local_ehi = procnode2.scheduler._local_ehi
    network_ehi = procnode2.scheduler._network_ehi
    for i, inst in enumerate(batch_const.instances):
        tasks = TaskGraphBuilder.from_program(
            batch_const.info.program,
            inst.pid,
            local_ehi,
            network_ehi,
            first_task_id=task_counter,
            prog_input=batch_const.info.inputs[i].values,
        )
        task_counter += len(tasks.get_tasks())
        tasks_const.append(tasks)
    procnode2.scheduler._task_from_block_builder._task_id_counter = task_counter

    # tasks_const = procnode2.scheduler.get_tasks_to_schedule_for(batch_const.batch_id)
    # merged_const = TaskGraphBuilder.merge_linear(tasks_const)
    merged_const = TaskGraphBuilder.merge(tasks_const)
    start = const_start
    for tid, tinfo in merged_const.get_tasks().items():
        merged_const.get_tinfo(tid).start_time = start
        if sched_typ == SchedulerType.FCFS:
            # Force the busy tasks to start earlier by setting their start times as deadlines
            merged_const.get_tinfo(tid).deadline = start
        start += const_period

    procnode2.scheduler.upload_task_graph(merged_const)

    # for tid, tinfo in merged_const.get_tasks().items():
    #     print(f"tid: {tid}, tinfo: {tinfo}")
    # if sched_typ == SchedulerType.NO_SCHED:
    #     merged_with_const = TaskGraphBuilder.merge_linear([merged2, merged_const])
    # else:
    #     merged_with_const = TaskGraphBuilder.merge([merged2, merged_const])

    # procnode2.scheduler.upload_task_graph(merged_with_const)

    # Run

    network.start()
    ns.sim_run()

    results: Dict[str, BatchResult] = {}
    statistics: Dict[str, SchedulerStatistics] = {}

    # only one batch (ID = 0), so get value at index 0
    results[node1] = procnode1.scheduler.get_batch_results()[0]
    results[node2] = procnode2.scheduler.get_batch_results()[0]
    statistics[node1] = procnode1.scheduler.get_statistics()
    statistics[node2] = procnode2.scheduler.get_statistics()

    total_duration = ns.sim_time()

    return AppResult(results, statistics, total_duration)


def run_two_node_app_separate_inputs_plus_local_program(
    num_iterations: int,
    num_local_iterations: int,
    node1: str,
    node2: str,
    prog_node1: QoalaProgram,
    prog_node1_inputs: List[ProgramInput],
    prog_node2: QoalaProgram,
    prog_node2_inputs: List[ProgramInput],
    local_prog_node2: QoalaProgram,
    local_prog_node2_inputs: List[ProgramInput],
    network_cfg: ProcNodeNetworkConfig,
    linear: bool = False,
    linear_for: Optional[Dict[str, bool]] = None,
    linear_local: bool = True,
) -> AppResult:
    if linear_for is None:
        linear_for = {node1: False, node2: False}

    ns.sim_reset()
    ns.set_qstate_formalism(ns.QFormalism.DM)
    seed = random.randint(0, 1000)
    ns.set_random_state(seed=seed)

    network = build_network_from_config(network_cfg)

    procnode1 = network.nodes[node1]
    unit_module1 = UnitModule.from_full_ehi(procnode1.memmgr.get_ehi())
    batch_info1 = create_batch(
        prog_node1, unit_module1, prog_node1_inputs, num_iterations
    )
    batch_node1 = procnode1.submit_batch(batch_info1)

    procnode2 = network.nodes[node2]
    unit_module2 = UnitModule.from_full_ehi(procnode2.memmgr.get_ehi())
    batch_info2 = create_batch(
        prog_node2, unit_module2, prog_node2_inputs, num_iterations
    )
    batch_node2 = procnode2.submit_batch(batch_info2)

    batch_info_local = create_batch(
        local_prog_node2, unit_module2, local_prog_node2_inputs, num_local_iterations
    )
    procnode2.submit_batch(batch_info_local)

    # Init

    remote_pids1 = {batch_node2.batch_id: [p.pid for p in batch_node2.instances]}
    procnode1.initialize_processes(remote_pids1)

    remote_pids2 = {batch_node1.batch_id: [p.pid for p in batch_node1.instances]}
    procnode2.initialize_processes(remote_pids2)

    network.start()
    ns.sim_run()

    results: Dict[str, BatchResult] = {}
    statistics: Dict[str, SchedulerStatistics] = {}

    # only one batch (ID = 0), so get value at index 0
    results[node1] = procnode1.scheduler.get_batch_results()[0]
    results[node2] = procnode2.scheduler.get_batch_results()[0]
    results["local"] = procnode2.scheduler.get_batch_results()[1]
    statistics[node1] = procnode1.scheduler.get_statistics()
    statistics[node2] = procnode2.scheduler.get_statistics()

    total_duration = ns.sim_time()

    return AppResult(results, statistics, total_duration)


def run_1_server_n_clients(
    client_names: List[str],
    client_program: QoalaProgram,
    server_name: str,
    server_program: QoalaProgram,
    client_inputs: Dict[str, List[ProgramInput]],
    server_inputs: List[ProgramInput],
    network_cfg: ProcNodeNetworkConfig,
    linear: bool = False,
) -> AppResult:
    ns.sim_reset()
    ns.set_qstate_formalism(ns.QFormalism.DM)
    seed = random.randint(0, 1000)
    ns.set_random_state(seed=seed)

    network = build_network_from_config(network_cfg)

    server_pids: Dict[str, int] = {}  # client name -> server PID

    for client_name in client_names:
        procnode = network.nodes[client_name]
        inputs = client_inputs[client_name]
        unit_module = UnitModule.from_full_ehi(procnode.memmgr.get_ehi())
        batch_info = create_batch(client_program, unit_module, inputs, 1)
        procnode.submit_batch(batch_info)

        server_procnode = network.nodes[server_name]
        server_unit_module = UnitModule.from_full_ehi(server_procnode.memmgr.get_ehi())
        program = deepcopy(server_program)
        program.meta.csockets[0] = client_name
        program.meta.epr_sockets[0] = client_name
        server_batch_info = create_batch(program, server_unit_module, server_inputs, 1)
        batch = server_procnode.submit_batch(server_batch_info)
        server_pids[client_name] = batch.instances[0].pid

    for client_name in client_names:
        procnode.initialize_processes({0: [server_pids[client_name]]}, linear=False)

    server_procnode.initialize_processes(
        {i: [0] for i in range(len(client_names))}, linear=False
    )

    network.start()
    ns.sim_run()

    results: Dict[str, BatchResult] = {}
    statistics: Dict[str, SchedulerStatistics] = {}

    for name in client_names:
        procnode = network.nodes[name]
        # only one batch (ID = 0), so get value at index 0
        results[name] = procnode.scheduler.get_batch_results()[0]
        statistics[name] = procnode.scheduler.get_statistics()

    server_procnode = network.nodes[server_name]
    results[server_name] = server_procnode.scheduler.get_batch_results()[0]
    statistics[server_name] = server_procnode.scheduler.get_statistics()

    total_duration = ns.sim_time()
    return AppResult(results, statistics, total_duration)


def run_two_node_app(
    num_iterations: int,
    programs: Dict[str, QoalaProgram],
    program_inputs: Dict[str, ProgramInput],
    network_cfg: ProcNodeNetworkConfig,
    linear: bool = False,
    linear_for: Optional[Dict[str, bool]] = None,
) -> AppResult:
    names = list(programs.keys())
    new_inputs = {
        name: [program_inputs[name] for _ in range(num_iterations)] for name in names
    }

    return run_two_node_app_separate_inputs(
        num_iterations, programs, new_inputs, network_cfg, linear, linear_for
    )


def run_single_node_app_separate_inputs(
    num_iterations: int,
    program_name: str,
    program: QoalaProgram,
    program_input: List[ProgramInput],
    network_cfg: ProcNodeNetworkConfig,
    linear: bool = False,
) -> AppResult:
    ns.sim_reset()
    ns.set_qstate_formalism(ns.QFormalism.DM)
    seed = random.randint(0, 1000)
    ns.set_random_state(seed=seed)

    network = build_network_from_config(network_cfg)

    procnode = list(network.nodes.values())[0]

    unit_module = UnitModule.from_full_ehi(procnode.memmgr.get_ehi())
    batch_info = create_batch(program, unit_module, program_input, num_iterations)
    procnode.submit_batch(batch_info)

    procnode.initialize_processes(linear=linear)

    # logger = LogManager.get_stack_logger()
    # for batch_id, prog_batch in procnode.scheduler.get_batches().items():
    #     task_graph = prog_batch.instances[0].task_graph
    #     num = len(prog_batch.instances)
    #     logger.info(f"batch {batch_id}: {num} instances each with task graph:")
    #     logger.info(task_graph)

    network.start()
    ns.sim_run()

    # only one batch (ID = 0), so get value at index 0
    results = procnode.scheduler.get_batch_results()[0]
    statistics = procnode.scheduler.get_statistics()
    total_duration = ns.sim_time()

    return AppResult(
        {program_name: results}, {program_name: statistics}, total_duration
    )


def run_single_node_app(
    num_iterations: int,
    program_name: str,
    program: QoalaProgram,
    program_input: ProgramInput,
    network_cfg: ProcNodeNetworkConfig,
    linear: bool = False,
) -> AppResult:
    new_inputs = [program_input for _ in range(num_iterations)]

    return run_single_node_app_separate_inputs(
        num_iterations, program_name, program, new_inputs, network_cfg, linear
    )



# Low-level Hardware Info. Expressed using NetSquid concepts and objects.
from __future__ import annotations

import itertools
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, FrozenSet, List, Optional, Type

from netsquid.components.instructions import (
    INSTR_CNOT,
    INSTR_CZ,
    INSTR_H,
    INSTR_INIT,
    INSTR_MEASURE,
    INSTR_ROT_X,
    INSTR_ROT_Y,
    INSTR_ROT_Z,
    INSTR_X,
    INSTR_Y,
    INSTR_Z,
    IMeasure,
)
from netsquid.components.instructions import Instruction as NetSquidInstruction
from netsquid.components.models.qerrormodels import (
    DepolarNoiseModel,
    QuantumErrorModel,
    T1T2NoiseModel,
)
from netsquid_magic.state_delivery_sampler import (
    DepolariseWithFailureStateSamplerFactory,
    IStateDeliverySamplerFactory,
    PerfectStateSamplerFactory,
)

from qoala.lang.common import MultiQubit
from qoala.runtime.instructions import (
    INSTR_BICHROMATIC,
    INSTR_MEASURE_ALL,
    INSTR_ROT_X_ALL,
    INSTR_ROT_Y_ALL,
    INSTR_ROT_Z_ALL,
)

# A measurement that is meant to take 0 time.
# Used in e.g. Measure Directly EPR generation.
INSTR_MEASURE_INSTANT = IMeasure("measurement_instant_op")


# Config Interface


class LhiQubitConfigInterface(ABC):
    @abstractmethod
    def to_is_communication(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def to_error_model(self) -> Type[QuantumErrorModel]:
        raise NotImplementedError

    @abstractmethod
    def to_error_model_kwargs(self) -> Dict[str, Any]:
        raise NotImplementedError


class LhiGateConfigInterface(ABC):
    @abstractmethod
    def to_instruction(self) -> Type[NetSquidInstruction]:
        raise NotImplementedError

    @abstractmethod
    def to_duration(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def to_error_model(self) -> Type[QuantumErrorModel]:
        raise NotImplementedError

    @abstractmethod
    def to_error_model_kwargs(self) -> Dict[str, Any]:
        raise NotImplementedError


class LhiTopologyConfigInterface(ABC):
    @abstractmethod
    def get_qubit_configs(self) -> Dict[int, LhiQubitConfigInterface]:
        raise NotImplementedError

    @abstractmethod
    def get_single_gate_configs(self) -> Dict[int, List[LhiGateConfigInterface]]:
        raise NotImplementedError

    @abstractmethod
    def get_multi_gate_configs(
        self,
    ) -> Dict[MultiQubit, List[LhiGateConfigInterface]]:
        raise NotImplementedError

    def get_all_qubit_gate_configs(self) -> List[LhiGateConfigInterface]:
        raise NotImplementedError


# Data classes


@dataclass
class LhiQubitInfo:
    is_communication: bool
    error_model: Type[QuantumErrorModel]
    error_model_kwargs: Dict[str, Any]


@dataclass
class LhiGateInfo:
    instruction: Type[NetSquidInstruction]
    duration: float  # ns
    error_model: Type[QuantumErrorModel]
    error_model_kwargs: Dict[str, Any]


@dataclass
class LhiTopology:
    qubit_infos: Dict[int, LhiQubitInfo]  # qubit ID -> info
    single_gate_infos: Dict[int, List[LhiGateInfo]]  # qubit ID -> gates
    multi_gate_infos: Dict[
        MultiQubit, List[LhiGateInfo]
    ]  # ordered qubit ID list -> gates
    all_qubit_gate_infos: Optional[List[LhiGateInfo]] = None

    def find_single_gate(
        self, qubit_id: int, instr: Type[NetSquidInstruction]
    ) -> Optional[LhiGateInfo]:
        if qubit_id not in self.single_gate_infos:
            return None
        for info in self.single_gate_infos[qubit_id]:
            if info.instruction == instr:
                return info
        return None

    def find_multi_gate(
        self, qubit_ids: List[int], instr: Type[NetSquidInstruction]
    ) -> Optional[LhiGateInfo]:
        multi = MultiQubit(qubit_ids)
        if multi not in self.multi_gate_infos:
            return None
        for info in self.multi_gate_infos[multi]:
            if info.instruction == instr:
                return info
        return None

    def find_all_qubit_gate(
        self, instr: Type[NetSquidInstruction]
    ) -> Optional[LhiGateInfo]:
        if self.all_qubit_gate_infos is None:
            return None

        for info in self.all_qubit_gate_infos:
            if info.instruction == instr:
                return info
        return None


# Convenience methods.


class LhiTopologyBuilder:
    """Convenience methods for creating a Topology object."""

    @classmethod
    def from_config(cls, cfg: LhiTopologyConfigInterface) -> LhiTopology:
        qubit_infos: Dict[int, LhiQubitInfo] = {}
        for i, cfg_info in cfg.get_qubit_configs().items():
            qubit_infos[i] = LhiQubitInfo(
                is_communication=cfg_info.to_is_communication(),
                error_model=cfg_info.to_error_model(),
                error_model_kwargs=cfg_info.to_error_model_kwargs(),
            )

        single_gate_infos: Dict[int, List[LhiGateInfo]] = {}
        for i, cfg_infos in cfg.get_single_gate_configs().items():
            single_gate_infos[i] = [
                LhiGateInfo(
                    instruction=info.to_instruction(),
                    duration=info.to_duration(),
                    error_model=info.to_error_model(),
                    error_model_kwargs=info.to_error_model_kwargs(),
                )
                for info in cfg_infos
            ]

        multi_gate_infos: Dict[MultiQubit, List[LhiGateInfo]] = {}
        for ids, cfg_infos in cfg.get_multi_gate_configs().items():
            multi_gate_infos[ids] = [
                LhiGateInfo(
                    instruction=info.to_instruction(),
                    duration=info.to_duration(),
                    error_model=info.to_error_model(),
                    error_model_kwargs=info.to_error_model_kwargs(),
                )
                for info in cfg_infos
            ]

        all_qubit_gate_infos: Optional[List[LhiGateInfo]] = None
        if cfg.get_all_qubit_gate_configs() is not None:
            all_qubit_gate_infos = []
            for info in cfg.get_all_qubit_gate_configs():
                all_qubit_gate_infos.append(
                    LhiGateInfo(
                        instruction=info.to_instruction(),
                        duration=info.to_duration(),
                        error_model=info.to_error_model(),
                        error_model_kwargs=info.to_error_model_kwargs(),
                    )
                )

        return LhiTopology(
            qubit_infos=qubit_infos,
            single_gate_infos=single_gate_infos,
            multi_gate_infos=multi_gate_infos,
            all_qubit_gate_infos=all_qubit_gate_infos,
        )

    @classmethod
    def t1t2_qubit(cls, is_communication: bool, t1: float, t2: float) -> LhiQubitInfo:
        return LhiQubitInfo(
            is_communication=is_communication,
            error_model=T1T2NoiseModel,
            error_model_kwargs={"T1": t1, "T2": t2},
        )

    @classmethod
    def perfect_qubit(cls, is_communication: bool) -> LhiQubitInfo:
        return cls.t1t2_qubit(is_communication=is_communication, t1=0, t2=0)

    @classmethod
    def depolar_gates(
        cls,
        duration: float,
        instructions: List[NetSquidInstruction],
        depolar_rate: float,
    ) -> List[LhiGateInfo]:
        return [
            LhiGateInfo(
                instruction=instr,
                duration=duration,
                error_model=DepolarNoiseModel,
                error_model_kwargs={"depolar_rate": depolar_rate},
            )
            for instr in instructions
        ]

    @classmethod
    def perfect_gates(
        cls, duration: float, instructions: List[NetSquidInstruction]
    ) -> List[LhiGateInfo]:
        return cls.depolar_gates(
            duration=duration, instructions=instructions, depolar_rate=0
        )

    @classmethod
    def perfect_uniform_default_gates(cls, num_qubits) -> LhiTopology:
        # TODO: test this and update default values
        return cls.perfect_uniform(
            num_qubits=num_qubits,
            single_instructions=[
                INSTR_INIT,
                INSTR_X,
                INSTR_Y,
                INSTR_Z,
                INSTR_H,
                INSTR_ROT_X,
                INSTR_ROT_Y,
                INSTR_ROT_Z,
                INSTR_MEASURE,
                INSTR_MEASURE_INSTANT,
            ],
            single_duration=5e3,
            two_instructions=[INSTR_CNOT, INSTR_CZ],
            two_duration=100e3,
        )

    @classmethod
    def perfect_uniform(
        cls,
        num_qubits,
        single_instructions: List[NetSquidInstruction],
        single_duration: float,
        two_instructions: List[NetSquidInstruction],
        two_duration: float,
        all_qubit_instructions: List[NetSquidInstruction] = None,
        all_qubit_duration: float = 0,
    ) -> LhiTopology:
        if all_qubit_instructions is None:
            return cls.fully_uniform(
                num_qubits=num_qubits,
                qubit_info=cls.perfect_qubit(is_communication=True),
                single_gate_infos=cls.perfect_gates(
                    single_duration, single_instructions
                ),
                two_gate_infos=cls.perfect_gates(two_duration, two_instructions),
            )
        else:
            all_qubit_gate_infos = cls.perfect_gates(
                all_qubit_duration, all_qubit_instructions
            )
            return cls.fully_uniform(
                num_qubits=num_qubits,
                qubit_info=cls.perfect_qubit(is_communication=True),
                single_gate_infos=cls.perfect_gates(
                    single_duration, single_instructions
                ),
                two_gate_infos=cls.perfect_gates(two_duration, two_instructions),
                all_qubit_gate_infos=all_qubit_gate_infos,
            )

    @classmethod
    def fully_uniform(
        cls,
        num_qubits,
        qubit_info: LhiQubitInfo,
        single_gate_infos: List[LhiGateInfo],
        two_gate_infos: List[LhiGateInfo],
        all_qubit_gate_infos: List[LhiGateInfo] = None,
    ) -> LhiTopology:
        q_infos = {i: qubit_info for i in range(num_qubits)}
        sg_infos = {i: single_gate_infos for i in range(num_qubits)}
        mg_infos = {}
        for i in range(num_qubits):
            for j in range(num_qubits):
                if i != j:
                    multi = MultiQubit([i, j])
                    mg_infos[multi] = two_gate_infos

        return LhiTopology(q_infos, sg_infos, mg_infos, all_qubit_gate_infos)

    @classmethod
    def perfect_star(
        cls,
        num_qubits: int,
        comm_instructions: List[NetSquidInstruction],
        comm_duration: float,
        mem_instructions: List[NetSquidInstruction],
        mem_duration: float,
        two_instructions: List[NetSquidInstruction],
        two_duration: float,
    ) -> LhiTopology:
        comm_qubit_info = cls.perfect_qubit(is_communication=True)
        mem_qubit_info = cls.perfect_qubit(is_communication=False)
        comm_gate_infos = cls.perfect_gates(comm_duration, comm_instructions)
        mem_gate_infos = cls.perfect_gates(mem_duration, mem_instructions)
        two_gate_infos = cls.perfect_gates(two_duration, two_instructions)

        q_infos = {0: comm_qubit_info}
        for i in range(1, num_qubits):
            q_infos[i] = mem_qubit_info

        sg_infos = {0: comm_gate_infos}
        for i in range(1, num_qubits):
            sg_infos[i] = mem_gate_infos

        mg_infos = {}
        for i in range(1, num_qubits):
            mg_infos[MultiQubit([0, i])] = two_gate_infos

        return LhiTopology(q_infos, sg_infos, mg_infos)

    @classmethod
    def generic_t1t2_star(
        cls,
        num_qubits: int,
        comm_t1: float,
        comm_t2: float,
        mem_t1: float,
        mem_t2: float,
        comm_instructions: List[NetSquidInstruction],
        comm_duration: float,
        comm_instr_depolar_rate: float,
        mem_instructions: List[NetSquidInstruction],
        mem_duration: float,
        mem_instr_depolar_rate: float,
        two_instructions: List[NetSquidInstruction],
        two_duration: float,
        two_instr_depolar_rate: float,
    ) -> LhiTopology:
        comm_qubit_info = cls.t1t2_qubit(is_communication=True, t1=comm_t1, t2=comm_t2)
        mem_qubit_info = cls.t1t2_qubit(is_communication=False, t1=mem_t1, t2=mem_t2)

        comm_gate_infos = cls.depolar_gates(
            comm_duration, comm_instructions, comm_instr_depolar_rate
        )
        mem_gate_infos = cls.depolar_gates(
            mem_duration, mem_instructions, mem_instr_depolar_rate
        )
        two_gate_infos = cls.depolar_gates(
            two_duration, two_instructions, two_instr_depolar_rate
        )

        q_infos = {0: comm_qubit_info}
        for i in range(1, num_qubits):
            q_infos[i] = mem_qubit_info

        sg_infos = {0: comm_gate_infos}
        for i in range(1, num_qubits):
            sg_infos[i] = mem_gate_infos

        mg_infos = {}
        for i in range(1, num_qubits):
            mg_infos[MultiQubit([0, i])] = two_gate_infos

        return LhiTopology(q_infos, sg_infos, mg_infos)

    @classmethod
    def trapped_ion_default_perfect_gates(cls, num_qubits: int) -> LhiTopology:
        # TODO check default values
        return cls.perfect_uniform(
            num_qubits=num_qubits,
            single_instructions=[
                INSTR_INIT,
                INSTR_ROT_Z,
                INSTR_MEASURE,
                INSTR_MEASURE_INSTANT,
            ],
            single_duration=5e3,
            two_instructions=[],
            two_duration=0,
            all_qubit_instructions=[
                INSTR_INIT,
                INSTR_MEASURE_ALL,
                INSTR_ROT_X_ALL,
                INSTR_ROT_Y_ALL,
                INSTR_ROT_Z_ALL,
                INSTR_BICHROMATIC,
            ],
            all_qubit_duration=5e3,
        )


class LhiLatenciesConfigInterface(ABC):
    @abstractmethod
    def get_host_instr_time(self) -> float:
        raise NotImplementedError

    @abstractmethod
    def get_qnos_instr_time(self) -> float:
        raise NotImplementedError

    @abstractmethod
    def get_host_peer_latency(self) -> float:
        raise NotImplementedError

    def get_internal_sched_latency(self):
        raise NotImplementedError


@dataclass
class LhiLatencies:
    host_instr_time: float = 0  # duration of classical Host instr execution
    qnos_instr_time: float = 0  # duration of classical Qnos instr execution
    host_peer_latency: float = 0  # processing time for Host messages from remote node
    internal_sched_latency: float = 0  # processing time for messaging between node scheduler and processor schedulers

    @classmethod
    def from_config(cls, cfg: LhiLatenciesConfigInterface) -> LhiLatencies:
        return LhiLatencies(
            host_instr_time=cfg.get_host_instr_time(),
            qnos_instr_time=cfg.get_qnos_instr_time(),
            host_peer_latency=cfg.get_host_peer_latency(),
            internal_sched_latency=cfg.get_internal_sched_latency(),
        )

    @classmethod
    def all_zero(cls) -> LhiLatencies:
        # NOTE: can also just use LhiLatencies() which will default all values to 0
        # However, using this classmethod makes this behavior more explicit and clear.
        return LhiLatencies(0, 0, 0, 0)


class LhiLinkConfigInterface(ABC):
    @abstractmethod
    def to_sampler_factory(self) -> Type[IStateDeliverySamplerFactory]:
        raise NotImplementedError

    @abstractmethod
    def to_sampler_kwargs(self) -> Dict[str, Any]:
        raise NotImplementedError

    @abstractmethod
    def to_state_delay(self) -> float:
        raise NotImplementedError


class LhiNetworkScheduleConfigInterface(ABC):
    @abstractmethod
    def to_bin_length(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def to_first_bin(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def to_bin_pattern(self) -> List[LhiNetworkTimebin]:
        raise NotImplementedError

    @abstractmethod
    def to_repeat_period(self) -> int:
        raise NotImplementedError


@dataclass
class LhiLinkInfo:
    sampler_factory: Type[IStateDeliverySamplerFactory]
    sampler_kwargs: Dict[str, Any]
    state_delay: float  # time between EPR generation and putting the state into memory

    @classmethod
    def from_config(cls, cfg: LhiLinkConfigInterface) -> LhiLinkInfo:
        return LhiLinkInfo(
            sampler_factory=cfg.to_sampler_factory(),
            sampler_kwargs=cfg.to_sampler_kwargs(),
            state_delay=cfg.to_state_delay(),
        )

    @classmethod
    def perfect(cls, duration: float) -> LhiLinkInfo:
        return LhiLinkInfo(
            sampler_factory=PerfectStateSamplerFactory,
            sampler_kwargs={"cycle_time": 0},
            state_delay=duration,
        )

    @classmethod
    def depolarise(
        cls,
        cycle_time: float,
        prob_max_mixed: float,
        prob_success: float,
        state_delay: float,
    ) -> LhiLinkInfo:
        return LhiLinkInfo(
            sampler_factory=DepolariseWithFailureStateSamplerFactory,
            sampler_kwargs={
                "cycle_time": cycle_time,
                "prob_max_mixed": prob_max_mixed,
                "prob_success": prob_success,
            },
            state_delay=state_delay,
        )


@dataclass
class LhiNetworkInfo:
    nodes: Dict[int, str]  # node ID -> node name

    # (node A ID, node B ID) -> link info
    # for a pair (a, b) there exists no separate (b, a) info (it is the same)
    links: Dict[FrozenSet[int], LhiLinkInfo]

    @classmethod
    def fully_connected(
        cls, nodes: Dict[int, str], info: LhiLinkInfo
    ) -> LhiNetworkInfo:
        links: Dict[FrozenSet[int], LhiLinkInfo] = {}
        for n1, n2 in itertools.combinations(nodes.keys(), 2):
            node_link = frozenset([n1, n2])
            links[node_link] = info
        return LhiNetworkInfo(nodes, links)

    @classmethod
    def perfect_fully_connected(
        cls, nodes: Dict[int, str], duration: float
    ) -> LhiNetworkInfo:
        link = LhiLinkInfo.perfect(duration)
        return cls.fully_connected(nodes, link)

    def add_link(self, node1_id, node2_id, link_info: LhiLinkInfo):
        if node1_id not in self.nodes:
            raise ValueError(f"Node with ID {node1_id} not found")
        if node2_id not in self.nodes:
            raise ValueError(f"Node with ID {node2_id} not found")
        if node1_id == node2_id:
            raise ValueError("Cannot add link between same node")
        node_link = frozenset([node1_id, node2_id])
        if node_link in self.links:
            raise ValueError(
                f"Link between nodes {node1_id} and {node2_id} already exists"
            )
        self.links[node_link] = link_info

    def get_link(self, node_id1: int, node_id2: int) -> LhiLinkInfo:
        node_link = frozenset([node_id1, node_id2])
        try:
            return self.links[node_link]
        except KeyError:
            raise ValueError(
                f"There is no link between nodes {node_id1} and {node_id2} in the network"
            )


@dataclass
class LhiNetworkTimebin:
    nodes: FrozenSet[int]
    pids: Dict[int, int]  # node ID -> PID


@dataclass
class LhiNetworkSchedule:
    bin_length: int
    first_bin: int
    bin_pattern: List[LhiNetworkTimebin]
    repeat_period: int

    @classmethod
    def from_config(
        cls, config: LhiNetworkScheduleConfigInterface
    ) -> LhiNetworkSchedule:
        return LhiNetworkSchedule(
            config.to_bin_length(),
            config.to_first_bin(),
            config.to_bin_pattern(),
            config.to_repeat_period(),
        )


@dataclass
class LhiProcNodeInfo:
    id: int
    name: str
    topology: LhiTopology
    latencies: LhiLatencies



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Generator, List, Optional, Tuple

import netsquid as ns
import pytest
from netqasm.lang.subroutine import Subroutine

from pydynaa import EventExpression
from qoala.lang.ehi import EhiBuilder, UnitModule
from qoala.lang.hostlang import (
    AddCValueOp,
    AssignCValueOp,
    BasicBlock,
    BasicBlockType,
    BitConditionalMultiplyConstantCValueOp,
    BusyOp,
    ClassicalIqoalaOp,
    IqoalaSingleton,
    IqoalaTuple,
    IqoalaVector,
    IqoalaVectorElement,
    MultiplyConstantCValueOp,
    ReceiveCMsgOp,
    ReturnResultOp,
    RunRequestOp,
    RunSubroutineOp,
    SendCMsgOp,
)
from qoala.lang.program import ProgramMeta, QoalaProgram
from qoala.lang.request import (
    CallbackType,
    EprRole,
    EprType,
    QoalaRequest,
    RequestRoutine,
    RequestVirtIdMapping,
)
from qoala.lang.routine import LocalRoutine, RoutineMetadata
from qoala.runtime.memory import ProgramMemory
from qoala.runtime.message import Message
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.runtime.sharedmem import SharedMemory
from qoala.sim.host.csocket import ClassicalSocket
from qoala.sim.host.hostinterface import HostInterface, HostLatencies
from qoala.sim.host.hostprocessor import HostProcessor
from qoala.sim.process import QoalaProcess
from qoala.util.tests import netsquid_run, yield_from

MOCK_MESSAGE = Message(src_pid=0, dst_pid=0, content=42)
MOCK_QNOS_RET_REG = "R0"
MOCK_QNOS_RET_VALUE = 7
MOCK_NETSTACK_RET_VALUE = 22


@dataclass(frozen=True)
class InterfaceEvent:
    peer: str
    msg: Message


class MockHostInterface(HostInterface):
    def __init__(self, shared_mem: Optional[SharedMemory] = None) -> None:
        self.send_events: List[InterfaceEvent] = []
        self.recv_events: List[InterfaceEvent] = []

        self.shared_mem = shared_mem
        self._program_instance_jumps: Dict[int, int] = {}  # pid => block name

    @property
    def program_instance_jumps(self) -> Dict[int, int]:
        return self._program_instance_jumps

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        self.send_events.append(InterfaceEvent(peer, msg))

    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
        self.recv_events.append(InterfaceEvent(peer, MOCK_MESSAGE))
        return MOCK_MESSAGE
        yield  # to make it behave as a generator

    def send_qnos_msg(self, msg: Message) -> None:
        self.send_events.append(InterfaceEvent("qnos", msg))

    def get_available_messages(self, peer: str) -> List[Tuple[int, int]]:
        return []

    def receive_qnos_msg(self) -> Generator[EventExpression, None, Message]:
        self.recv_events.append(InterfaceEvent("qnos", MOCK_MESSAGE))
        if self.shared_mem is not None:
            # Hack to find out which addr was allocated to write results to.
            result_addr = self.shared_mem._lr_out_addrs[0]
            self.shared_mem.write_lr_out(result_addr, [MOCK_QNOS_RET_VALUE])
        return MOCK_MESSAGE
        yield  # to make it behave as a generator

    @property
    def name(self) -> str:
        return "mock"


def create_program(
    instrs: Optional[List[ClassicalIqoalaOp]] = None,
    subroutines: Optional[Dict[str, LocalRoutine]] = None,
    requests: Optional[Dict[str, RequestRoutine]] = None,
    meta: Optional[ProgramMeta] = None,
) -> QoalaProgram:
    if instrs is None:
        instrs = []

    if meta is None:
        meta = ProgramMeta.empty("prog")
    # TODO: split into proper blocks
    block = BasicBlock("b0", BasicBlockType.CL, instrs)
    return QoalaProgram(
        blocks=[block], local_routines=subroutines, request_routines=requests, meta=meta
    )


def create_process(
    program: QoalaProgram,
    interface: HostInterface,
    inputs: Optional[Dict[str, Any]] = None,
) -> QoalaProcess:
    if inputs is None:
        inputs = {}
    prog_input = ProgramInput(values=inputs)

    mock_ehi = EhiBuilder.perfect_uniform(
        num_qubits=2,
        flavour=None,
        single_instructions=[],
        single_duration=0,
        two_instructions=[],
        two_duration=0,
    )

    instance = ProgramInstance(
        pid=0,
        program=program,
        inputs=prog_input,
        unit_module=UnitModule.from_full_ehi(mock_ehi),
    )

    mem = ProgramMemory(pid=0)
    process = QoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={
            id: ClassicalSocket(interface, name, 0, 0)
            for (id, name) in program.meta.csockets.items()
        },
        epr_sockets=program.meta.epr_sockets,
        result=ProgramResult(values={}),
    )
    return process


def test_initialize():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())
    program = create_program()
    process = create_process(
        program, interface, inputs={"x": 1, "theta": 3.14, "name": "alice"}
    )

    processor.initialize(process)
    host_mem = process.prog_memory.host_mem

    assert host_mem.read("x") == 1
    assert host_mem.read("theta") == 3.14
    assert host_mem.read("name") == "alice"

    with pytest.raises(KeyError):
        host_mem.read_vec("x")


def test_assign_cvalue():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())
    program = create_program(instrs=[AssignCValueOp(IqoalaSingleton("x"), 3)])
    process = create_process(program, interface)
    processor.initialize(process)

    yield_from(processor.assign_instr_index(process, 0))
    assert process.prog_memory.host_mem.read("x") == 3


def test_assign_cvalue_with_latencies():
    ns.sim_reset()

    interface = MockHostInterface()
    latencies = HostLatencies(host_instr_time=500)
    processor = HostProcessor(interface, latencies)
    program = create_program(instrs=[AssignCValueOp(IqoalaSingleton("x"), 3)])
    process = create_process(program, interface)
    processor.initialize(process)

    assert ns.sim_time() == 0
    netsquid_run(processor.assign_instr_index(process, 0))
    assert process.prog_memory.host_mem.read("x") == 3
    assert ns.sim_time() == 500


def test_busy():
    ns.sim_reset()

    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())
    program = create_program(instrs=[BusyOp(500)])
    process = create_process(program, interface)
    processor.initialize(process)

    assert ns.sim_time() == 0
    netsquid_run(processor.assign_instr_index(process, 0))
    assert ns.sim_time() == 500


def test_busy_with_latencies():
    ns.sim_reset()

    interface = MockHostInterface()
    latencies = HostLatencies(host_instr_time=1000, host_peer_latency=3000)
    processor = HostProcessor(interface, latencies)
    program = create_program(instrs=[BusyOp(500)])
    process = create_process(program, interface)
    processor.initialize(process)

    assert ns.sim_time() == 0
    netsquid_run(processor.assign_instr_index(process, 0))
    assert ns.sim_time() == 500  # latencies are not important for busy op


def test_send_msg():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())
    meta = ProgramMeta.empty("alice")
    meta.csockets = {0: "bob"}

    program = create_program(
        instrs=[SendCMsgOp(IqoalaSingleton("bob"), IqoalaSingleton("msg"))], meta=meta
    )
    process = create_process(program, interface, inputs={"bob": 0, "msg": 12})
    processor.initialize(process)

    yield_from(processor.assign_instr_index(process, 0))
    assert interface.send_events[0] == InterfaceEvent("bob", Message(0, 0, content=12))


def test_send_msg_with_latencies():
    ns.sim_reset()

    interface = MockHostInterface()
    latencies = HostLatencies(host_instr_time=500, host_peer_latency=1e6)
    processor = HostProcessor(interface, latencies)
    meta = ProgramMeta.empty("alice")
    meta.csockets = {0: "bob"}
    program = create_program(
        instrs=[SendCMsgOp(IqoalaSingleton("bob"), IqoalaSingleton("msg"))], meta=meta
    )
    process = create_process(program, interface, inputs={"bob": 0, "msg": 12})
    processor.initialize(process)

    assert ns.sim_time() == 0
    netsquid_run(processor.assign_instr_index(process, 0))
    assert interface.send_events[0] == InterfaceEvent("bob", Message(0, 0, content=12))
    assert ns.sim_time() == 500  # no host_peer_latency used!


def test_recv_msg():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())
    meta = ProgramMeta.empty("alice")
    meta.csockets = {0: "bob"}
    program = create_program(
        instrs=[ReceiveCMsgOp(IqoalaSingleton("bob"), IqoalaSingleton("msg"))],
        meta=meta,
    )
    process = create_process(program, interface, inputs={"bob": 0})
    processor.initialize(process)

    yield_from(processor.assign_instr_index(process, 0))
    assert interface.recv_events[0] == InterfaceEvent("bob", MOCK_MESSAGE)
    assert process.prog_memory.host_mem.read("msg") == MOCK_MESSAGE.content


def test_recv_msg_with_latencies():
    ns.sim_reset()

    interface = MockHostInterface()
    latencies = HostLatencies(host_instr_time=500, host_peer_latency=1e6)
    processor = HostProcessor(interface, latencies)
    meta = ProgramMeta.empty("alice")
    meta.csockets = {0: "bob"}
    program = create_program(
        instrs=[ReceiveCMsgOp(IqoalaSingleton("bob"), IqoalaSingleton("msg"))],
        meta=meta,
    )
    process = create_process(program, interface, inputs={"bob": 0})
    processor.initialize(process)

    assert ns.sim_time() == 0
    netsquid_run(processor.assign_instr_index(process, 0))
    assert interface.recv_events[0] == InterfaceEvent("bob", MOCK_MESSAGE)
    assert process.prog_memory.host_mem.read("msg") == MOCK_MESSAGE.content
    assert ns.sim_time() == 1e6  # no host_instr_time used !


def test_add_cvalue():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())
    program = create_program(
        instrs=[
            AssignCValueOp(IqoalaSingleton("a"), 2),
            AssignCValueOp(IqoalaSingleton("b"), 3),
            AddCValueOp(
                IqoalaSingleton("sum"), IqoalaSingleton("a"), IqoalaSingleton("b")
            ),
        ]
    )
    process = create_process(program, interface)
    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign_instr_index(process, i))

    assert process.prog_memory.host_mem.read("sum") == 5


def test_add_cvalue_with_inputs():
    ns.sim_reset()

    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies(host_instr_time=500))
    program = create_program(
        instrs=[
            AddCValueOp(
                IqoalaSingleton("sum"), IqoalaSingleton("a"), IqoalaSingleton("b")
            ),
        ]
    )
    process = create_process(
        program,
        interface,
        inputs={"a": 2, "b": 3},
    )
    processor.initialize(process)

    assert ns.sim_time() == 0
    netsquid_run(processor.assign_instr_index(process, 0))

    assert process.prog_memory.host_mem.read("sum") == 5
    assert ns.sim_time() == 500


def test_add_cvalue_with_latencies():
    ns.sim_reset()

    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies(host_instr_time=1200))
    program = create_program(
        instrs=[
            AssignCValueOp(IqoalaSingleton("a"), 2),
            AssignCValueOp(IqoalaSingleton("b"), 3),
            AddCValueOp(
                IqoalaSingleton("sum"), IqoalaSingleton("a"), IqoalaSingleton("b")
            ),
        ]
    )
    process = create_process(program, interface)
    processor.initialize(process)

    assert ns.sim_time() == 0
    for i in range(len(program.instructions)):
        netsquid_run(processor.assign_instr_index(process, i))

    assert process.prog_memory.host_mem.read("sum") == 5
    assert ns.sim_time() == len(program.instructions) * 1200


def test_multiply_const():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())
    program = create_program(
        instrs=[
            AssignCValueOp(IqoalaSingleton("a"), 4),
            MultiplyConstantCValueOp(
                IqoalaSingleton("result"), IqoalaSingleton("a"), -1
            ),
        ]
    )
    process = create_process(program, interface)
    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign_instr_index(process, i))

    assert process.prog_memory.host_mem.read("result") == -4


def test_multiply_const_with_inputs():
    ns.sim_reset()

    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies(host_instr_time=500))
    program = create_program(
        instrs=[
            MultiplyConstantCValueOp(
                IqoalaSingleton("result"), IqoalaSingleton("a"), -1
            )
        ]
    )
    process = create_process(program, interface, inputs={"a": 4})
    processor.initialize(process)

    for i in range(len(program.instructions)):
        netsquid_run(processor.assign_instr_index(process, i))

    assert process.prog_memory.host_mem.read("result") == -4
    assert ns.sim_time() == len(program.instructions) * 500


def test_multiply_const_with_latencies():
    ns.sim_reset()

    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies(host_instr_time=500))
    program = create_program(
        instrs=[
            AssignCValueOp(IqoalaSingleton("a"), 4),
            MultiplyConstantCValueOp(
                IqoalaSingleton("result"), IqoalaSingleton("a"), -1
            ),
        ]
    )
    process = create_process(program, interface)
    processor.initialize(process)

    for i in range(len(program.instructions)):
        netsquid_run(processor.assign_instr_index(process, i))

    assert process.prog_memory.host_mem.read("result") == -4
    assert ns.sim_time() == len(program.instructions) * 500


def test_bit_cond_mult():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())
    program = create_program(
        instrs=[
            AssignCValueOp(IqoalaSingleton("var1"), 4),
            AssignCValueOp(IqoalaSingleton("var2"), 7),
            AssignCValueOp(IqoalaSingleton("cond1"), 0),
            AssignCValueOp(IqoalaSingleton("cond2"), 1),
            BitConditionalMultiplyConstantCValueOp(
                IqoalaSingleton("result1"),
                IqoalaSingleton("var1"),
                IqoalaSingleton("cond1"),
                -1,
            ),
            BitConditionalMultiplyConstantCValueOp(
                IqoalaSingleton("result2"),
                IqoalaSingleton("var2"),
                IqoalaSingleton("cond2"),
                -1,
            ),
        ]
    )
    process = create_process(program, interface)
    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign_instr_index(process, i))

    assert process.prog_memory.host_mem.read("result1") == 4
    assert process.prog_memory.host_mem.read("result2") == -7


def test_bit_cond_mult_with_inputs():
    ns.sim_reset()
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies(host_instr_time=500))
    program = create_program(
        instrs=[
            AssignCValueOp(IqoalaSingleton("var2"), 7),
            BitConditionalMultiplyConstantCValueOp(
                IqoalaSingleton("result1"),
                IqoalaSingleton("var1"),
                IqoalaSingleton("cond1"),
                -1,
            ),
            BitConditionalMultiplyConstantCValueOp(
                IqoalaSingleton("result2"),
                IqoalaSingleton("var2"),
                IqoalaSingleton("cond2"),
                -1,
            ),
        ]
    )
    process = create_process(
        program, interface, inputs={"var1": 4, "cond1": 0, "cond2": 1}
    )
    processor.initialize(process)

    assert ns.sim_time() == 0
    for i in range(len(program.instructions)):
        netsquid_run(processor.assign_instr_index(process, i))

    assert process.prog_memory.host_mem.read("result1") == 4
    assert process.prog_memory.host_mem.read("result2") == -7
    assert ns.sim_time() == len(program.instructions) * 500


def test_bit_cond_mult_with_latencies():
    ns.sim_reset()
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies(host_instr_time=500))
    program = create_program(
        instrs=[
            AssignCValueOp(IqoalaSingleton("var1"), 4),
            AssignCValueOp(IqoalaSingleton("var2"), 7),
            AssignCValueOp(IqoalaSingleton("cond1"), 0),
            AssignCValueOp(IqoalaSingleton("cond2"), 1),
            BitConditionalMultiplyConstantCValueOp(
                IqoalaSingleton("result1"),
                IqoalaSingleton("var1"),
                IqoalaSingleton("cond1"),
                -1,
            ),
            BitConditionalMultiplyConstantCValueOp(
                IqoalaSingleton("result2"),
                IqoalaSingleton("var2"),
                IqoalaSingleton("cond2"),
                -1,
            ),
        ]
    )
    process = create_process(program, interface)
    processor.initialize(process)

    assert ns.sim_time() == 0
    for i in range(len(program.instructions)):
        netsquid_run(processor.assign_instr_index(process, i))

    assert process.prog_memory.host_mem.read("result1") == 4
    assert process.prog_memory.host_mem.read("result2") == -7
    assert ns.sim_time() == len(program.instructions) * 500


def test_prepare_lr_call():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())

    subrt = Subroutine()
    metadata = RoutineMetadata.use_none()
    routine = LocalRoutine(
        "subrt1", subrt, return_vars=[IqoalaVector("res", 3)], metadata=metadata
    )
    instr = RunSubroutineOp(
        result=IqoalaVector("res", 3), values=IqoalaTuple([]), subrt="subrt1"
    )

    program = create_program(instrs=[instr], subroutines={"subrt1": routine})
    process = create_process(program, interface)
    processor.initialize(process)

    lrcall = processor.prepare_lr_call(process, program.instructions[0])

    # Host processor should have allocated shared memory space.
    assert lrcall.routine_name == "subrt1"
    assert len(process.shared_mem.raw_arrays.raw_memory[lrcall.result_addr]) == 3


def test_post_lr_call():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())

    subrt = Subroutine()
    metadata = RoutineMetadata.use_none()
    routine = LocalRoutine(
        "subrt1", subrt, return_vars=[IqoalaVector("res", 3)], metadata=metadata
    )
    instr = RunSubroutineOp(
        result=IqoalaVector("res", 3), values=IqoalaTuple([]), subrt="subrt1"
    )

    program = create_program(instrs=[instr], subroutines={"subrt1": routine})
    process = create_process(program, interface)
    processor.initialize(process)

    lrcall = processor.prepare_lr_call(process, program.instructions[0])
    # Mock LR execution by writing results to shared memory.
    process.shared_mem.write_lr_out(lrcall.result_addr, [1, 2, 3])

    processor.post_lr_call(process, program.instructions[0], lrcall)

    # Host memory should contain the results.
    assert process.host_mem.read_vec("res") == [1, 2, 3]


def create_simple_request(
    remote_id: int,
    num_pairs: int,
    virt_ids: RequestVirtIdMapping,
    typ: EprType,
    role: EprRole,
) -> QoalaRequest:
    return QoalaRequest(
        name="req",
        remote_id=remote_id,
        epr_socket_id=0,
        num_pairs=num_pairs,
        virt_ids=virt_ids,
        timeout=1000,
        fidelity=0.65,
        typ=typ,
        role=role,
    )


def test_prepare_rr_call():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())

    request = create_simple_request(
        remote_id=0,
        num_pairs=10,
        virt_ids=RequestVirtIdMapping.from_str("increment 0"),
        typ=EprType.MEASURE_DIRECTLY,
        role=EprRole.CREATE,
    )
    routine = RequestRoutine(
        "req", request, [IqoalaVector("outcomes", 10)], CallbackType.WAIT_ALL, None
    )
    instr = RunRequestOp(
        result=IqoalaVector("outcomes", 10), values=IqoalaTuple([]), routine="req"
    )

    program = create_program(instrs=[instr], requests={"req": routine})
    process = create_process(program, interface)
    processor.initialize(process)

    rrcall = processor.prepare_rr_call(process, program.instructions[0])

    # Host processor should have allocated shared memory space.
    assert rrcall.routine_name == "req"
    assert len(process.shared_mem.raw_arrays.raw_memory[rrcall.result_addr]) == 10
    assert rrcall.cb_input_addrs == []
    assert rrcall.cb_output_addrs == []


def test_prepare_rr_call_with_callbacks_seq():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())

    subrt = Subroutine()
    metadata = RoutineMetadata.use_none()
    local_routine = LocalRoutine(
        "subrt1", subrt, return_vars=[IqoalaVector("res", 3)], metadata=metadata
    )

    request = create_simple_request(
        remote_id=0,
        num_pairs=10,
        virt_ids=RequestVirtIdMapping.from_str("increment 0"),
        typ=EprType.MEASURE_DIRECTLY,
        role=EprRole.CREATE,
    )
    routine = RequestRoutine("req", request, [], CallbackType.SEQUENTIAL, "subrt1")

    # We expect 10 (num_pairs) * 3 (per callback) = 30 results
    instr = RunRequestOp(
        result=IqoalaVector("outcomes", 30), values=IqoalaTuple([]), routine="req"
    )

    program = create_program(
        instrs=[instr], subroutines={"subrt1": local_routine}, requests={"req": routine}
    )
    process = create_process(program, interface)
    processor.initialize(process)

    rrcall = processor.prepare_rr_call(process, program.instructions[0])

    # Host processor should have allocated shared memory space.
    assert rrcall.routine_name == "req"
    raw_memory = process.shared_mem.raw_arrays.raw_memory
    # 0 entries for the RR itself
    assert len(raw_memory[rrcall.result_addr]) == 0
    # 3 entries for each of the callbacks
    assert all(len(raw_memory[addr]) == 3 for addr in rrcall.cb_output_addrs)

    assert len(process.shared_mem._cr_in_addrs) == 10

    assert rrcall.cb_input_addrs == process.shared_mem._cr_in_addrs

    # 1 address for each of the callbacks in total 10 addresses
    assert len(rrcall.cb_input_addrs) == 10
    assert len(rrcall.cb_output_addrs) == 10


def test_prepare_rr_call_with_callbacks_wait_all():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())

    subrt = Subroutine()
    metadata = RoutineMetadata.use_none()
    local_routine = LocalRoutine(
        "subrt1", subrt, return_vars=[IqoalaVector("res", 3)], metadata=metadata
    )

    request = create_simple_request(
        remote_id=0,
        num_pairs=10,
        virt_ids=RequestVirtIdMapping.from_str("increment 0"),
        typ=EprType.MEASURE_DIRECTLY,
        role=EprRole.CREATE,
    )
    routine = RequestRoutine(
        "req",
        request,
        [IqoalaVector("req_return", 2)],
        CallbackType.WAIT_ALL,
        "subrt1",
    )

    # We expect 3 (callback) + 2 (request) = 5 results
    instr = RunRequestOp(
        result=IqoalaVector("outcomes", 5), values=IqoalaTuple([]), routine="req"
    )

    program = create_program(
        instrs=[instr], subroutines={"subrt1": local_routine}, requests={"req": routine}
    )
    process = create_process(program, interface)
    processor.initialize(process)

    rrcall = processor.prepare_rr_call(process, program.instructions[0])

    # Host processor should have allocated shared memory space.
    assert rrcall.routine_name == "req"
    raw_memory = process.shared_mem.raw_arrays.raw_memory
    # 2 entries for the RR itself
    assert len(raw_memory[rrcall.result_addr]) == 2
    # 3 entries for each of the callbacks
    assert all(len(raw_memory[addr]) == 3 for addr in rrcall.cb_output_addrs)

    assert len(process.shared_mem._cr_in_addrs) == 1
    assert rrcall.cb_input_addrs == process.shared_mem._cr_in_addrs

    # just one call of the callback
    assert len(rrcall.cb_input_addrs) == 1
    assert len(rrcall.cb_output_addrs) == 1


def test_post_rr_call_with_callbacks_seq():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())

    subrt = Subroutine()
    metadata = RoutineMetadata.use_none()
    local_routine = LocalRoutine(
        "subrt1", subrt, return_vars=[IqoalaVector("res", 3)], metadata=metadata
    )

    request = create_simple_request(
        remote_id=0,
        num_pairs=10,
        virt_ids=RequestVirtIdMapping.from_str("increment 0"),
        typ=EprType.MEASURE_DIRECTLY,
        role=EprRole.CREATE,
    )
    routine = RequestRoutine("req", request, [], CallbackType.SEQUENTIAL, "subrt1")

    # We expect 10 (num_pairs) * 3 (per callback) = 30 results
    instr = RunRequestOp(
        result=IqoalaVector("outcomes", 30), values=IqoalaTuple([]), routine="req"
    )

    program = create_program(
        instrs=[instr], subroutines={"subrt1": local_routine}, requests={"req": routine}
    )
    process = create_process(program, interface)
    processor.initialize(process)

    rrcall = processor.prepare_rr_call(process, program.instructions[0])
    # Mock RR execution by writing results to shared memory.
    for i in range(10):
        data = [3 * i, 3 * i + 1, 3 * i + 2]
        process.shared_mem.write_lr_out(rrcall.cb_output_addrs[i], data)

    processor.post_rr_call(process, program.instructions[0], rrcall)

    # Host memory should contain the results.
    assert process.host_mem.read_vec("outcomes") == [i for i in range(30)]


def test_post_rr_call_with_callbacks_wait_all():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())

    subrt = Subroutine()
    metadata = RoutineMetadata.use_none()
    local_routine = LocalRoutine(
        "subrt1", subrt, return_vars=[IqoalaVector("res", 3)], metadata=metadata
    )

    request = create_simple_request(
        remote_id=0,
        num_pairs=10,
        virt_ids=RequestVirtIdMapping.from_str("increment 0"),
        typ=EprType.MEASURE_DIRECTLY,
        role=EprRole.CREATE,
    )
    routine = RequestRoutine(
        "req",
        request,
        [IqoalaVector("req_result", 2)],
        CallbackType.WAIT_ALL,
        "subrt1",
    )

    # We expect 3 (callback) + 2 (request) = 5 results
    instr = RunRequestOp(
        result=IqoalaVector("outcomes", 5), values=IqoalaTuple([]), routine="req"
    )

    program = create_program(
        instrs=[instr], subroutines={"subrt1": local_routine}, requests={"req": routine}
    )
    process = create_process(program, interface)
    processor.initialize(process)

    rrcall = processor.prepare_rr_call(process, program.instructions[0])
    # Mock RR execution by writing results to shared memory.
    process.shared_mem.write_rr_out(rrcall.result_addr, [0, 1])
    data = [2, 3, 4]
    process.shared_mem.write_lr_out(rrcall.cb_output_addrs[0], data)

    processor.post_rr_call(process, program.instructions[0], rrcall)

    # Host memory should contain the results.
    assert process.host_mem.read_vec("outcomes") == [i for i in range(5)]


def test_return_result():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())
    program = create_program(
        instrs=[
            AssignCValueOp(IqoalaSingleton("result"), 2),
            ReturnResultOp(IqoalaSingleton("result")),
        ]
    )
    process = create_process(program, interface)
    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign_instr_index(process, i))

    assert process.prog_memory.host_mem.read("result") == 2
    assert process.result.values == {"result": 2}


def test_vector_element():
    interface = MockHostInterface()
    processor = HostProcessor(interface, HostLatencies.all_zero())

    subrt = Subroutine()
    metadata = RoutineMetadata.use_none()
    routine = LocalRoutine(
        "subrt1", subrt, return_vars=[IqoalaVector("res", 3)], metadata=metadata
    )
    instr = RunSubroutineOp(
        result=IqoalaVector("res", 3), values=IqoalaTuple([]), subrt="subrt1"
    )
    instr2 = AddCValueOp(
        IqoalaSingleton("result"),
        IqoalaVectorElement("res", 0),
        IqoalaVectorElement("res", 1),
    )

    program = create_program(instrs=[instr, instr2], subroutines={"subrt1": routine})
    process = create_process(program, interface)
    processor.initialize(process)

    lrcall = processor.prepare_lr_call(process, program.instructions[0])
    # Mock LR execution by writing results to shared memory.
    process.shared_mem.write_lr_out(lrcall.result_addr, [1, 2, 3])

    processor.post_lr_call(process, program.instructions[0], lrcall)

    yield_from(processor.assign_instr_index(process, 1))

    # Host memory should contain the results.
    assert process.host_mem.read_vec("res") == [1, 2, 3]
    assert process.prog_memory.host_mem.read("result") == 3


if __name__ == "__main__":
    test_initialize()
    test_assign_cvalue()
    test_assign_cvalue_with_latencies()
    test_busy()
    test_busy_with_latencies()
    test_send_msg()
    test_send_msg_with_latencies()
    test_recv_msg()
    test_recv_msg_with_latencies()
    test_add_cvalue()
    test_add_cvalue_with_inputs()
    test_add_cvalue_with_latencies()
    test_multiply_const()
    test_multiply_const_with_inputs()
    test_multiply_const_with_latencies()
    test_bit_cond_mult()
    test_bit_cond_mult_with_inputs()
    test_bit_cond_mult_with_latencies()
    test_prepare_lr_call()
    test_post_lr_call()
    test_prepare_rr_call()
    test_prepare_rr_call_with_callbacks_seq()
    test_prepare_rr_call_with_callbacks_wait_all()
    test_post_rr_call_with_callbacks_seq()
    test_post_rr_call_with_callbacks_wait_all()
    test_return_result()
    test_vector_element()



import math
from typing import List

import numpy as np
from netsquid.qubits import ketstates, qubitapi
from netsquid.qubits.qubit import Qubit

B00_DENS = np.outer(ketstates.b00, ketstates.b00)
B01_DENS = np.outer(ketstates.b01, ketstates.b01)
B10_DENS = np.outer(ketstates.b10, ketstates.b10)
B11_DENS = np.outer(ketstates.b11, ketstates.b11)

S00_DENS = np.outer(ketstates.s00, ketstates.s00)
S01_DENS = np.outer(ketstates.s01, ketstates.s01)
S10_DENS = np.outer(ketstates.s10, ketstates.s10)
S11_DENS = np.outer(ketstates.s11, ketstates.s11)

ONE_MAX_MIXED = np.array([[0.5, 0], [0, 0.5]])
TWO_MAX_MIXED = np.array(
    [
        [0.25, 0, 0, 0],
        [0, 0.25, 0, 0],
        [0, 0, 0.25, 0],
        [0, 0, 0, 0.25],
    ]
)

PI = math.pi
PI_OVER_2 = math.pi / 2


def fidelity_to_prob_max_mixed(num_qubits: int, fid: float) -> float:
    if num_qubits == 1:
        return (1 - fid) * 2.0
    elif num_qubits == 2:
        return (1 - fid) * 4.0 / 3.0
    else:
        raise ValueError


def prob_max_mixed_to_fidelity(num_qubits: int, prob: float) -> float:
    if num_qubits == 1:
        return 1 - 0.5 * prob
    elif num_qubits == 2:
        return 1 - 0.75 * prob
    else:
        raise ValueError


def has_state(qubit: Qubit, state: np.ndarray, margin: float = 0.001) -> bool:
    fidelity = qubitapi.fidelity(qubit, state, squared=True)
    return math.isclose(1.0, fidelity, abs_tol=margin)


def has_multi_state(
    qubits: List[Qubit], state: np.ndarray, margin: float = 0.001
) -> bool:
    fidelity = qubitapi.fidelity(qubits, state, squared=True)
    return math.isclose(1.0, fidelity, abs_tol=margin)


def density_matrices_equal(
    state1: np.ndarray, state2: np.ndarray, margin: float = 0.001
) -> bool:
    distance = np.linalg.norm(state1 - state2, 1) / 2
    return math.isclose(0, distance, abs_tol=margin)


def has_max_mixed_state(qubit: Qubit, margin: float = 0.001) -> bool:
    max_mixed = np.array([[0.5, 0], [0, 0.5]])
    qubit_state = qubitapi.reduced_dm(qubit)
    return density_matrices_equal(qubit_state, max_mixed, margin)


def has_max_mixed_two_qubit_state(q0: Qubit, q1: Qubit, margin: float = 0.001) -> bool:
    max_mixed = TWO_MAX_MIXED
    qubit_state = qubitapi.reduced_dm([q0, q1])
    return density_matrices_equal(qubit_state, max_mixed, margin)



from __future__ import annotations

import json
import math
import os
import time
from argparse import ArgumentParser
from dataclasses import asdict, dataclass
from datetime import datetime
from pathlib import Path
from typing import List

import netsquid as ns
import numpy as np

from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    ClassicalConnectionConfig,
    LatenciesConfig,
    NetworkScheduleConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchResult, ProgramInput
from qoala.util.runner import run_two_node_app_separate_inputs


def create_procnode_cfg(
    name: str,
    id: int,
    t1: float,
    t2: float,
    determ: bool,
    deadlines: bool,
    num_qubits: int,
    prio_epr: bool,
) -> ProcNodeConfig:
    topology = TopologyConfig.uniform_t1t2_qubits_perfect_gates_default_params(
        num_qubits, t1, t2
    )
    if name == "alice":
        latencies = LatenciesConfig(qnos_instr_time=1000, host_instr_time=1000)
    elif name == "bob":
        latencies = LatenciesConfig(qnos_instr_time=100_000, host_instr_time=1000)

    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=topology,
        latencies=latencies,
        ntf=NtfConfig.from_cls_name("GenericNtf"),
        determ_sched=determ,
        use_deadlines=deadlines,
        prio_epr=prio_epr,
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


@dataclass
class TeleportResult:
    alice_result: BatchResult
    bob_result: BatchResult
    total_duration: float


def run_apps(
    num_iterations: int,
    num_qubits_bob: int,
    t1: float,
    t2: float,
    cc_latency: float,
    network_bin_len: int,
    network_period: int,
    network_first_bin: int,
    prio_epr: bool,
) -> TeleportResult:
    ns.sim_reset()

    alice_id = 1
    bob_id = 0

    alice_node_cfg = create_procnode_cfg(
        "alice",
        alice_id,
        t1,
        t2,
        determ=True,
        deadlines=True,
        num_qubits=20,
        prio_epr=prio_epr,
    )
    bob_node_cfg = create_procnode_cfg(
        "bob",
        bob_id,
        t1,
        t2,
        determ=True,
        deadlines=True,
        num_qubits=num_qubits_bob,
        prio_epr=prio_epr,
    )

    cconn = ClassicalConnectionConfig.from_nodes(alice_id, bob_id, cc_latency)
    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[alice_node_cfg, bob_node_cfg], link_duration=1000
    )
    pattern = [(alice_id, i, bob_id, i) for i in range(num_iterations)]
    network_cfg.netschedule = NetworkScheduleConfig(
        bin_length=network_bin_len,
        first_bin=network_first_bin,
        bin_pattern=pattern,
        repeat_period=network_period,
    )
    network_cfg.cconns = [cconn]

    alice_program = load_program("programs/teleport_alice.iqoala")
    bob_program = load_program("programs/teleport_bob.iqoala")

    alice_inputs = [
        ProgramInput({"bob_id": bob_id, "state": i % 6}) for i in range(num_iterations)
    ]
    bob_inputs = [
        ProgramInput({"alice_id": alice_id, "state": i % 6})
        for i in range(num_iterations)
    ]

    app_result = run_two_node_app_separate_inputs(
        num_iterations=num_iterations,
        programs={"alice": alice_program, "bob": bob_program},
        program_inputs={"alice": alice_inputs, "bob": bob_inputs},
        network_cfg=network_cfg,
        linear=False,
    )

    alice_result = app_result.batch_results["alice"]
    bob_result = app_result.batch_results["bob"]
    # print(local_result)

    return TeleportResult(alice_result, bob_result, app_result.total_duration)


@dataclass
class DataPoint:
    cc_latency: float
    num_qubits_bob: int
    tel_succ_prob: float
    tel_succ_prob_lower: float
    tel_succ_prob_upper: float
    makespan: float


@dataclass
class DataMeta:
    timestamp: str
    num_iterations: int
    latency_factor: float
    net_bin_factors: List[float]
    prio_epr: bool
    num_qubits: int
    num_steps: int
    t1: float
    t2: float


@dataclass
class Data:
    meta: DataMeta
    data_points: List[DataPoint]


def relative_to_cwd(file: str) -> str:
    return os.path.join(os.path.dirname(__file__), file)


def wilson_score_interval(p_hat, n, z):
    denominator = 1 + z**2 / n
    centre_adjusted_probability = p_hat + z**2 / (2 * n)
    adjusted_standard_deviation = z * math.sqrt(
        (p_hat * (1 - p_hat) + z**2 / (4 * n)) / n
    )

    lower_bound = (
        centre_adjusted_probability - adjusted_standard_deviation
    ) / denominator
    upper_bound = (
        centre_adjusted_probability + adjusted_standard_deviation
    ) / denominator

    return (lower_bound, upper_bound)


def get_datapoint(
    num_runs: int,
    num_iterations: int,
    num_qubits_bob: int,
    t1: float,
    t2: float,
    latency_factor: float,
    network_bin_len: int,
    network_period: int,
    network_first_bin: int,
    prio_epr: bool,
) -> DataPoint:
    teleport_successes: List[bool] = []
    makespans: List[float] = []

    for _ in range(num_runs):
        cc_latency = latency_factor * t2

        result = run_apps(
            num_iterations=num_iterations,
            num_qubits_bob=num_qubits_bob,
            t1=t1,
            t2=t2,
            cc_latency=cc_latency,
            network_period=network_period,
            network_bin_len=network_bin_len,
            network_first_bin=network_first_bin,
            prio_epr=prio_epr,
        )
        teleport_results = result.bob_result.results
        outcomes = [result.values["outcome"] for result in teleport_results]
        assert len(outcomes) == num_iterations
        teleport_successes.extend([outcomes[i] == 0 for i in range(num_iterations)])

        makespans.append(result.total_duration)

    tel_avg_succ_prob = sum([s for s in teleport_successes if s]) / len(
        teleport_successes
    )
    tel_succ_prob_lower, tel_succ_prob_upper = wilson_score_interval(
        p_hat=tel_avg_succ_prob, n=len(teleport_successes), z=1.96
    )
    tel_lower_rounded = round(tel_succ_prob_lower, 3)
    tel_upper_rounded = round(tel_succ_prob_upper, 3)
    tel_succprob_rounded = round(tel_avg_succ_prob, 3)

    makespan = sum(makespans) / len(makespans)

    print(
        f"teleport succ prob: {tel_succprob_rounded} ({tel_lower_rounded}, {tel_upper_rounded})"
    )
    print(f"makespan: {makespan:_}")

    return DataPoint(
        cc_latency=cc_latency,
        num_qubits_bob=num_qubits_bob,
        tel_succ_prob=tel_avg_succ_prob,
        tel_succ_prob_lower=tel_succ_prob_lower,
        tel_succ_prob_upper=tel_succ_prob_upper,
        makespan=makespan,
    )


def run(output_dir: str, num_runs: int, num_qubits: int, num_steps: int, dump: bool):
    start_time = time.time()
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")

    t1 = 1e10
    t2 = 1e8
    prio_epr = False

    num_iterations = 5
    latency_factor = 0.1  # CC = 10_000_000
    net_bin_factors = list(np.linspace(0.2, 0.6, num_steps + 1))
    net_bin_factors = [round(f, 2) for f in net_bin_factors]
    network_gap = 0

    data_points: List[DataPoint] = []

    for bin_factor in net_bin_factors:
        network_bin_len = int((latency_factor * t2) * bin_factor)
        network_first_bin = int(network_bin_len / 2)
        network_period = (num_iterations + network_gap) * network_bin_len

        data_point = get_datapoint(
            num_runs=num_runs,
            num_iterations=num_iterations,
            num_qubits_bob=num_qubits,
            t1=t1,
            t2=t2,
            latency_factor=latency_factor,
            network_bin_len=network_bin_len,
            network_period=network_period,
            network_first_bin=network_first_bin,
            prio_epr=prio_epr,
        )
        makespan = data_point.makespan
        num_bins = math.ceil(makespan / network_bin_len)
        # print(f"total duration: {end_time - start_time}s")

        print(f"cc latency: {latency_factor * t2:_}")
        print(f"network period: {network_period:_}")
        print(f"network bin len: {network_bin_len:_}")
        print(f"makespan: {makespan}")
        print(f"num bins: {num_bins}")

        data_points.append(data_point)

    meta = DataMeta(
        timestamp=timestamp,
        num_iterations=num_iterations,
        latency_factor=latency_factor,
        net_bin_factors=net_bin_factors,
        prio_epr=prio_epr,
        num_qubits=num_qubits,
        num_steps=num_steps,
        t1=t1,
        t2=t2,
    )

    end_time = time.time()
    print(f"simulation took {end_time - start_time}s")

    if dump:
        data = Data(meta=meta, data_points=data_points)
        json_data = asdict(data)

        abs_dir = relative_to_cwd(f"data/{output_dir}")
        Path(abs_dir).mkdir(parents=True, exist_ok=True)
        last_path = os.path.join(abs_dir, "LAST.json")
        timestamp_path = os.path.join(abs_dir, f"{timestamp}.json")
        with open(last_path, "w") as datafile:
            json.dump(json_data, datafile)
        with open(timestamp_path, "w") as datafile:
            json.dump(json_data, datafile)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--num_qubits", "-q", type=int, required=True)
    parser.add_argument("--num_steps", "-s", type=int, required=True)
    parser.add_argument("--num_runs", "-n", type=int, required=True)
    parser.add_argument("--dump", "-d", action="store_true")

    args = parser.parse_args()

    dump = args.dump
    num_qubits = args.num_qubits
    num_steps = args.num_steps
    num_runs = args.num_runs

    run(
        f"sweep_bin_length_{num_qubits}_{num_steps}",
        num_runs=num_runs,
        num_qubits=num_qubits,
        num_steps=num_steps,
        dump=dump,
    )



from typing import Dict, Generator

import netsquid as ns

from pydynaa import EventExpression
from qoala.runtime.task import ProcessorType, QoalaTask, TaskGraph
from qoala.sim.driver import Driver
from qoala.sim.events import EVENT_WAIT
from qoala.sim.scheduler import CpuEdfScheduler, Status
from qoala.util.logging import LogManager


class SimpleTask(QoalaTask):
    def __init__(self, task_id: int, duration: int) -> None:
        super().__init__(task_id, ProcessorType.CPU, 0, duration)


class MockDriver(Driver):
    def __init__(self) -> None:
        # Task ID -> timestamp of start
        self._executed_tasks: Dict[int, float] = {}

    def wait(self, delta_time: float) -> Generator[EventExpression, None, None]:
        self._schedule_after(delta_time, EVENT_WAIT)
        event_expr = EventExpression(source=self, event_type=EVENT_WAIT)
        yield event_expr

    def handle_task(self, task: QoalaTask) -> Generator[EventExpression, None, None]:
        now = ns.sim_time()
        self._executed_tasks[task.task_id] = now
        yield from self.wait(task.duration)
        return True
        yield


def test_update_status_one_root():
    graph = TaskGraph()
    graph.add_tasks([SimpleTask(0, 200), SimpleTask(1, 500)])
    graph.add_precedences([(0, 1)])
    graph.add_rel_deadlines([((0, 1), 100)])
    graph.get_tasks()

    scheduler = CpuEdfScheduler("sched", 0, MockDriver(), None, None)
    scheduler.add_tasks(graph.get_tasks())
    scheduler.update_status()
    assert scheduler.status.status == {Status.NEXT_TASK}
    assert scheduler.status.params == {"task_id": 0}


def test_update_status_two_roots():
    graph = TaskGraph()
    graph.add_tasks([SimpleTask(0, 200), SimpleTask(1, 500)])
    graph.add_deadlines([(0, 1000), (1, 500)])

    scheduler = CpuEdfScheduler("sched", 0, MockDriver(), None, None)
    scheduler.add_tasks(graph.get_tasks())
    scheduler.update_status()
    assert scheduler.status.status == {Status.NEXT_TASK}
    assert scheduler.status.params == {"task_id": 1}


def test_edf_1():
    LogManager.set_task_log_level("INFO")
    graph = TaskGraph()
    graph.add_tasks([SimpleTask(0, 200), SimpleTask(1, 500)])
    graph.add_precedences([(0, 1)])
    graph.add_rel_deadlines([((0, 1), 100)])

    scheduler = CpuEdfScheduler("sched", 0, MockDriver(), None, None)
    scheduler.add_tasks(graph.get_tasks())

    ns.sim_reset()
    scheduler.start()
    ns.sim_run()

    assert scheduler._driver._executed_tasks == {0: 0, 1: 200}


def test_edf_2():
    graph = TaskGraph()
    graph.add_tasks(
        [SimpleTask(1, 500), SimpleTask(2, 80), SimpleTask(3, 300), SimpleTask(4, 100)]
    )
    graph.add_precedences([(1, 2), (1, 3), (2, 4)])
    graph.add_rel_deadlines([((1, 2), 200), ((1, 3), 400), ((2, 4), 100)])

    scheduler = CpuEdfScheduler("sched", 0, MockDriver(), None, None)
    scheduler.add_tasks(graph.get_tasks())

    ns.sim_reset()
    scheduler.start()
    ns.sim_run()

    assert scheduler._driver._executed_tasks == {1: 0, 2: 500, 4: 580, 3: 680}


if __name__ == "__main__":
    test_update_status_one_root()
    test_update_status_two_roots()
    test_edf_1()
    test_edf_2()



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional, Union

from netqasm.lang.subroutine import Subroutine

from qoala.lang.hostlang import IqoalaVector


@dataclass
class RoutineMetadata:
    # IDs in unit module of virtual qubits that are
    # used in this routine
    qubit_use: List[int]

    # IDs in unit module of virtual qubits that still have a state
    # that should be kept after finishing this routine
    qubit_keep: List[int]

    @classmethod
    def use_none(cls) -> RoutineMetadata:
        return RoutineMetadata([], [])

    @classmethod
    def free_all(cls, ids: List[int]) -> RoutineMetadata:
        return RoutineMetadata(ids, [])


@dataclass(frozen=True)
class LocalRoutine:
    name: str
    subroutine: Subroutine
    return_vars: List[Union[str, IqoalaVector]]
    metadata: RoutineMetadata
    request_name: Optional[str] = None

    def get_return_size(self) -> int:
        size = 0
        for v in self.return_vars:
            if isinstance(v, IqoalaVector):
                assert isinstance(v.size, int)
                size += v.size
            else:
                size += 1
        return size

    def serialize(self) -> str:
        s = f"SUBROUTINE {self.name}"
        s += f"\nparams: {', '.join(self.subroutine.arguments)}"
        s += f"\nreturns: {', '.join(str(v) for v in self.return_vars)}"
        s += f"\nuses: {', '.join(str(q) for q in self.metadata.qubit_use)}"
        s += f"\nkeeps: {', '.join(str(q) for q in self.metadata.qubit_keep)}"
        s += "\nNETQASM_START\n"
        s += self.subroutine.print_instructions()
        s += "\nNETQASM_END"
        return s

    def __str__(self) -> str:
        s = "\n"
        for value in self.return_vars:
            s += f"return {str(value)}\n"
        s += "NETQASM_START\n"
        s += self.subroutine.print_instructions()
        s += "\nNETQASM_END"
        return s



from __future__ import annotations

import os
import random
from dataclasses import dataclass
from typing import Dict

import netsquid as ns

from qoala.lang.ehi import UnitModule
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    ClassicalConnectionConfig,
    LatenciesConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchResult, ProgramBatch, ProgramInput
from qoala.runtime.statistics import SchedulerStatistics
from qoala.sim.build import build_network_from_config
from qoala.util.runner import AppResult, create_batch


def create_procnode_cfg(
    name: str, id: int, num_qubits: int, determ: bool
) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=TopologyConfig.perfect_config_uniform_default_params(num_qubits),
        latencies=LatenciesConfig(qnos_instr_time=1000),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
        determ_sched=determ,
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


@dataclass
class GhzResult:
    alice_results: BatchResult
    bob_results: BatchResult
    charlie_results: BatchResult


def run_ghz(num_iterations: int) -> GhzResult:
    ns.sim_reset()

    num_qubits = 4
    alice_id = 0
    bob_id = 1
    charlie_id = 2

    alice_node_cfg = create_procnode_cfg("alice", alice_id, num_qubits, determ=True)
    bob_node_cfg = create_procnode_cfg("bob", bob_id, num_qubits, determ=True)
    charlie_node_cfg = create_procnode_cfg(
        "charlie", charlie_id, num_qubits, determ=True
    )

    cconn_ab = ClassicalConnectionConfig.from_nodes(alice_id, bob_id, 1e9)
    cconn_ac = ClassicalConnectionConfig.from_nodes(alice_id, charlie_id, 1e9)
    cconn_bc = ClassicalConnectionConfig.from_nodes(bob_id, charlie_id, 1e9)
    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[alice_node_cfg, bob_node_cfg, charlie_node_cfg], link_duration=1000
    )
    network_cfg.cconns = [cconn_ab, cconn_ac, cconn_bc]

    alice_program = load_program("ghz_alice.iqoala")
    bob_program = load_program("ghz_bob.iqoala")
    charlie_program = load_program("ghz_charlie.iqoala")

    alice_input = [
        ProgramInput({"bob_id": bob_id, "charlie_id": charlie_id})
        for _ in range(num_iterations)
    ]
    bob_input = [
        ProgramInput({"alice_id": alice_id, "charlie_id": charlie_id})
        for _ in range(num_iterations)
    ]
    charlie_input = [
        ProgramInput({"alice_id": alice_id, "bob_id": bob_id})
        for _ in range(num_iterations)
    ]

    ns.sim_reset()
    ns.set_qstate_formalism(ns.QFormalism.DM)
    seed = random.randint(0, 1000)
    ns.set_random_state(seed=seed)

    network = build_network_from_config(network_cfg)
    names = ["alice", "bob", "charlie"]
    programs = {"alice": alice_program, "bob": bob_program, "charlie": charlie_program}
    program_inputs = {"alice": alice_input, "bob": bob_input, "charlie": charlie_input}
    batches: Dict[str, ProgramBatch] = {}  # node -> batch

    for name in names:
        procnode = network.nodes[name]
        program = programs[name]
        inputs = program_inputs[name]

        unit_module = UnitModule.from_full_ehi(procnode.memmgr.get_ehi())
        batch_info = create_batch(program, unit_module, inputs, num_iterations)
        batches[name] = procnode.submit_batch(batch_info)

    for batch in batches.values():
        assert batch.batch_id == 0
        assert [p.pid for p in batch.instances] == [i for i in range(num_iterations)]

    for name in names:
        procnode = network.nodes[name]
        remote_pids = {0: [i for i in range(num_iterations)]}
        procnode.initialize_processes(remote_pids=remote_pids, linear=True)

    network.start()
    ns.sim_run()

    results: Dict[str, BatchResult] = {}
    statistics: Dict[str, SchedulerStatistics] = {}

    for name in names:
        procnode = network.nodes[name]
        # only one batch (ID = 0), so get value at index 0
        results[name] = procnode.scheduler.get_batch_results()[0]
        statistics[name] = procnode.scheduler.get_statistics()

    total_duration = ns.sim_time()
    app_result = AppResult(results, statistics, total_duration)

    alice_result = app_result.batch_results["alice"]
    bob_result = app_result.batch_results["bob"]
    charlie_result = app_result.batch_results["charlie"]

    return GhzResult(alice_result, bob_result, charlie_result)


def test_ghz():
    result = run_ghz(num_iterations=10)

    alice_results = result.alice_results.results
    alice_outcomes = [result.values["outcome"] for result in alice_results]
    print(alice_outcomes)

    bob_results = result.bob_results.results
    bob_outcomes = [result.values["outcome"] for result in bob_results]
    print(bob_outcomes)

    charlie_results = result.charlie_results.results
    charlie_outcomes = [result.values["outcome"] for result in charlie_results]
    print(charlie_outcomes)

    assert alice_outcomes == bob_outcomes == charlie_outcomes


if __name__ == "__main__":
    test_ghz()



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import List

import netsquid as ns

from qoala.lang.ehi import UnitModule
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    LatenciesConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchInfo, BatchResult, ProgramInput
from qoala.util.runner import run_two_node_app


def create_procnode_cfg(name: str, id: int, num_qubits: int) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=TopologyConfig.perfect_config_uniform_default_params(num_qubits),
        latencies=LatenciesConfig(qnos_instr_time=1000),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


def create_batch(
    program: QoalaProgram,
    unit_module: UnitModule,
    inputs: List[ProgramInput],
    num_iterations: int,
) -> BatchInfo:
    return BatchInfo(
        program=program,
        unit_module=unit_module,
        inputs=inputs,
        num_iterations=num_iterations,
        deadline=0,
    )


@dataclass
class BqcResult:
    client_results: BatchResult
    server_results: BatchResult


def run_bqc(
    alpha,
    beta,
    theta1,
    theta2,
    num_iterations: int,
) -> BqcResult:
    ns.sim_reset()

    num_qubits = 3
    client_id = 0
    server_id = 1

    server_node_cfg = create_procnode_cfg("server", server_id, num_qubits)
    client_node_cfg = create_procnode_cfg("client", client_id, num_qubits)

    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[server_node_cfg, client_node_cfg], link_duration=1000
    )

    server_program = load_program("bqc_server.iqoala")
    server_input = ProgramInput({"client_id": client_id})

    client_program = load_program("bqc_client.iqoala")
    client_input = ProgramInput(
        {
            "server_id": server_id,
            "alpha": alpha,
            "beta": beta,
            "theta1": theta1,
            "theta2": theta2,
        }
    )

    app_result = run_two_node_app(
        num_iterations=num_iterations,
        programs={"server": server_program, "client": client_program},
        program_inputs={"server": server_input, "client": client_input},
        network_cfg=network_cfg,
        linear=True,
    )

    client_results = app_result.batch_results["client"]
    server_results = app_result.batch_results["server"]

    return BqcResult(client_results, server_results)


def check(alpha, beta, theta1, theta2, expected, num_iterations):
    # Effective computation: measure in Z the following state:
    # H Rz(beta) H Rz(alpha) |+>
    # m2 should be this outcome

    # angles are in multiples of pi/16

    ns.sim_reset()
    bqc_result = run_bqc(
        alpha=alpha,
        beta=beta,
        theta1=theta1,
        theta2=theta2,
        num_iterations=num_iterations,
    )
    assert len(bqc_result.client_results.results) > 0
    assert len(bqc_result.server_results.results) > 0

    program_results = bqc_result.server_results.results
    m2s = [result.values["m2"] for result in program_results]
    assert all(m2 == expected for m2 in m2s)


def test_bqc():
    # LogManager.enable_task_logger(True)
    # LogManager.log_tasks_to_file("bqc_comp_1.log")
    check(alpha=8, beta=8, theta1=0, theta2=0, expected=0, num_iterations=10)
    check(alpha=8, beta=24, theta1=0, theta2=0, expected=1, num_iterations=10)
    check(alpha=8, beta=8, theta1=13, theta2=27, expected=0, num_iterations=10)
    check(alpha=8, beta=24, theta1=2, theta2=22, expected=1, num_iterations=10)


if __name__ == "__main__":
    test_bqc()
